# SPDX-License-Identifier: MIT
from dataclasses import dataclass, field
from typing import Any
from xml.etree import ElementTree

from .additionalaudience import AdditionalAudience
from .nameditemlist import NamedItemList
from .odxdoccontext import OdxDocContext
from .odxlink import OdxLinkDatabase, OdxLinkId, OdxLinkRef
from .odxtypes import odxstr_to_bool
from .snrefcontext import SnRefContext


@dataclass(kw_only=True)
class Audience:
    enabled_audience_refs: list[OdxLinkRef] = field(default_factory=list)
    disabled_audience_refs: list[OdxLinkRef] = field(default_factory=list)

    is_supplier_raw: bool | None = None
    is_development_raw: bool | None = None
    is_manufacturing_raw: bool | None = None
    is_aftersales_raw: bool | None = None
    is_aftermarket_raw: bool | None = None

    @property
    def is_supplier(self) -> bool:
        return self.is_supplier_raw in [None, True]

    @property
    def is_development(self) -> bool:
        return self.is_development_raw in [None, True]

    @property
    def is_manufacturing(self) -> bool:
        return self.is_manufacturing_raw in [None, True]

    @property
    def is_aftersales(self) -> bool:
        return self.is_aftersales_raw in [None, True]

    @property
    def is_aftermarket(self) -> bool:
        return self.is_aftermarket_raw in [None, True]

    @property
    def enabled_audiences(self) -> NamedItemList[AdditionalAudience]:
        return self._enabled_audiences

    @property
    def disabled_audiences(self) -> NamedItemList[AdditionalAudience]:
        return self._disabled_audiences

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "Audience":

        enabled_audience_refs = [
            OdxLinkRef.from_et(ref, context) for ref in et_element.iterfind("ENABLED-AUDIENCE-REFS/"
                                                                            "ENABLED-AUDIENCE-REF")
        ]
        disabled_audience_refs = [
            OdxLinkRef.from_et(ref, context)
            for ref in et_element.iterfind("DISABLED-AUDIENCE-REFS/"
                                           "DISABLED-AUDIENCE-REF")
        ]
        is_supplier_raw = odxstr_to_bool(et_element.attrib.get("IS-SUPPLIER"))
        is_development_raw = odxstr_to_bool(et_element.attrib.get("IS-DEVELOPMENT"))
        is_manufacturing_raw = odxstr_to_bool(et_element.attrib.get("IS-MANUFACTURING"))
        is_aftersales_raw = odxstr_to_bool(et_element.attrib.get("IS-AFTERSALES"))
        is_aftermarket_raw = odxstr_to_bool(et_element.attrib.get("IS-AFTERMARKET"))

        return Audience(
            enabled_audience_refs=enabled_audience_refs,
            disabled_audience_refs=disabled_audience_refs,
            is_supplier_raw=is_supplier_raw,
            is_development_raw=is_development_raw,
            is_manufacturing_raw=is_manufacturing_raw,
            is_aftersales_raw=is_aftersales_raw,
            is_aftermarket_raw=is_aftermarket_raw,
        )

    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        return {}

    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        self._enabled_audiences = NamedItemList(
            [odxlinks.resolve(ref, AdditionalAudience) for ref in self.enabled_audience_refs])
        self._disabled_audiences = NamedItemList(
            [odxlinks.resolve(ref, AdditionalAudience) for ref in self.disabled_audience_refs])

    def _resolve_snrefs(self, context: SnRefContext) -> None:
        pass



# SPDX-License-Identifier: MIT
from dataclasses import dataclass, field
from xml.etree import ElementTree

from .basicstructure import BasicStructure
from .exceptions import odxrequire
from .odxdoccontext import OdxDocContext
from .utils import dataclass_fields_asdict


@dataclass(kw_only=True)
class EnvironmentData(BasicStructure):
    """This class represents Environment Data that describes the
    circumstances in which the error occurred.

    This is one of the many multiplexer mechanisms specified by the
    ODX standard, because an environment data parameter must only be
    used if a DTC parameter has a certain set of values. (In this
    sense, it is quite similar to NRC-CONST parameters.)
    """

    all_value: bool | None = None
    dtc_values: list[int] = field(default_factory=list)

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "EnvironmentData":
        """Reads Environment Data from Diag Layer."""
        kwargs = dataclass_fields_asdict(BasicStructure.from_et(et_element, context))

        all_value_elem = et_element.find("ALL-VALUE")
        all_value = None if all_value_elem is None else True
        dtc_values = [
            int(odxrequire(dtcv_elem.text))
            for dtcv_elem in et_element.iterfind("DTC-VALUES/DTC-VALUE")
        ]

        return EnvironmentData(all_value=all_value, dtc_values=dtc_values, **kwargs)



# SPDX-License-Identifier: MIT
import unittest
from xml.etree import ElementTree

from packaging.version import Version

import odxtools_porsche.uds as uds
from odxtools.compumethods.compucategory import CompuCategory
from odxtools.compumethods.compuinternaltophys import CompuInternalToPhys
from odxtools.compumethods.compurationalcoeffs import CompuRationalCoeffs
from odxtools.compumethods.compuscale import CompuScale
from odxtools.compumethods.identicalcompumethod import IdenticalCompuMethod
from odxtools.compumethods.linearcompumethod import LinearCompuMethod
from odxtools.createanydiagcodedtype import create_any_diag_coded_type_from_et
from odxtools.database import Database
from odxtools.dataobjectproperty import DataObjectProperty
from odxtools.decodestate import DecodeState
from odxtools.description import Description
from odxtools.diagdatadictionaryspec import DiagDataDictionarySpec
from odxtools.diaglayers.diaglayertype import DiagLayerType
from odxtools.diaglayers.ecuvariant import EcuVariant
from odxtools.diaglayers.ecuvariantraw import EcuVariantRaw
from odxtools.encodestate import EncodeState
from odxtools.encoding import Encoding
from odxtools.exceptions import DecodeError, EncodeError, OdxError, odxrequire
from odxtools.leadinglengthinfotype import LeadingLengthInfoType
from odxtools.minmaxlengthtype import MinMaxLengthType
from odxtools.nameditemlist import NamedItemList
from odxtools.odxdoccontext import OdxDocContext
from odxtools.odxlink import DocType, OdxDocFragment, OdxLinkDatabase, OdxLinkId, OdxLinkRef
from odxtools.odxtypes import DataType
from odxtools.parameters.codedconstparameter import CodedConstParameter
from odxtools.parameters.lengthkeyparameter import LengthKeyParameter
from odxtools.parameters.valueparameter import ValueParameter
from odxtools.paramlengthinfotype import ParamLengthInfoType
from odxtools.physicaltype import PhysicalType
from odxtools.request import Request
from odxtools.standardlengthtype import StandardLengthType
from odxtools.termination import Termination

doc_frags = [OdxDocFragment("UnitTest", DocType.CONTAINER)]


class TestLeadingLengthInfoType(unittest.TestCase):

    def test_decode_leading_length_info_type_bytefield(self) -> None:
        dct = LeadingLengthInfoType(
            base_data_type=DataType.A_BYTEFIELD,
            bit_length=6,
        )
        state = DecodeState(bytes([0x2, 0x34, 0x56]), cursor_byte_position=0)
        internal_value = dct.decode_from_pdu(state)
        self.assertEqual(internal_value, bytes([0x34, 0x56]))

        dct = LeadingLengthInfoType(
            base_data_type=DataType.A_BYTEFIELD,
            bit_length=5,
        )
        # 0xC2 = 0b11000010, with bit_position=1 and bit_lenth=5, the extracted bits are 00001,
        # i.e. the leading length is 1, i.e. only the byte 0x3 should be extracted.
        state = DecodeState(
            bytes([0x1, 0xC2, 0x3, 0x4]), cursor_byte_position=1, cursor_bit_position=1)
        internal_value = dct.decode_from_pdu(state)
        self.assertEqual(internal_value, bytes([0x3]))

    def test_decode_leading_length_info_type_zero_length(self) -> None:
        dct = LeadingLengthInfoType(
            base_data_type=DataType.A_BYTEFIELD,
            bit_length=8,
        )
        state = DecodeState(bytes([0x0, 0x1]), cursor_byte_position=0)
        internal_value = dct.decode_from_pdu(state)
        self.assertEqual(internal_value, b'')
        self.assertEqual(state.cursor_byte_position, 1)

    def test_encode_leading_length_info_type_bytefield(self) -> None:
        dct = LeadingLengthInfoType(
            base_data_type=DataType.A_BYTEFIELD,
            bit_length=5,
        )
        state = EncodeState(bytearray(), cursor_bit_position=3)
        dct.encode_into_pdu(bytes([0x3]), state)
        self.assertEqual(state.coded_message.hex(), "0803")

    def test_decode_leading_length_info_type_bytefield2(self) -> None:
        dct = LeadingLengthInfoType(
            base_data_type=DataType.A_BYTEFIELD,
            bit_length=5,
        )

        state = EncodeState(
            coded_message=bytearray.fromhex("0000ff00"),
            used_mask=bytearray.fromhex("0700ffff"),
            cursor_bit_position=3)
        dct.encode_into_pdu(bytes([0xcc]), state)
        self.assertEqual(state.coded_message.hex(), "08ccff00")
        self.assertEqual(state.cursor_byte_position, 2)
        self.assertEqual(state.cursor_bit_position, 0)

    def test_decode_leading_length_info_type_unicode2string(self) -> None:
        # big endian
        dct = LeadingLengthInfoType(
            base_data_type=DataType.A_UNICODE2STRING,
            bit_length=7,
        )
        state = DecodeState(
            bytes([0x12, 0x8, 0x00, 0x61, 0x00, 0x39, 0xff, 0x00]),
            cursor_byte_position=1,
            cursor_bit_position=1)
        internal_value = dct.decode_from_pdu(state)
        self.assertEqual(internal_value, "a9")
        self.assertEqual(state.cursor_byte_position, 6)
        self.assertEqual(state.cursor_bit_position, 0)

        # little endian
        dct = LeadingLengthInfoType(
            base_data_type=DataType.A_UNICODE2STRING,
            bit_length=8,
            is_highlow_byte_order_raw=False,
        )
        state = DecodeState(
            bytes([0x12, 0x4, 0x61, 0x00, 0x39, 0x00, 0xff, 0x00]), cursor_byte_position=1)
        internal_value = dct.decode_from_pdu(state)
        self.assertEqual(internal_value, "a9")
        self.assertEqual(state.cursor_byte_position, 6)
        self.assertEqual(state.cursor_bit_position, 0)

    def test_encode_leading_length_info_type_unicode2string(self) -> None:
        dct = LeadingLengthInfoType(
            base_data_type=DataType.A_UNICODE2STRING,
            bit_length=8,
        )
        state = EncodeState(coded_message=bytearray())
        dct.encode_into_pdu("a9", state)
        self.assertEqual(state.coded_message, bytes([0x4, 0x00, 0x61, 0x00, 0x39]))

        dct = LeadingLengthInfoType(
            base_data_type=DataType.A_UNICODE2STRING,
            bit_length=8,
            is_highlow_byte_order_raw=False,
        )
        state = EncodeState(
            coded_message=bytearray(),
            cursor_byte_position=0,
            cursor_bit_position=0,
            origin_byte_position=0)
        dct.encode_into_pdu("a9", state)
        self.assertEqual(state.coded_message, bytes([0x4, 0x61, 0x00, 0x39, 0x00]))

    def test_end_to_end(self) -> None:
        # diag coded types
        diagcodedtypes = {
            "uint8":
                StandardLengthType(
                    base_data_type=DataType.A_UINT32,
                    bit_length=8,
                ),
            "certificateClient":
                LeadingLengthInfoType(
                    base_data_type=DataType.A_BYTEFIELD,
                    bit_length=8,
                ),
        }

        # computation methods
        compumethods = {
            "uint_passthrough":
                IdenticalCompuMethod(
                    category=CompuCategory.IDENTICAL,
                    internal_type=DataType.A_UINT32,
                    physical_type=DataType.A_UINT32),
            "bytes_passthrough":
                IdenticalCompuMethod(
                    category=CompuCategory.IDENTICAL,
                    internal_type=DataType.A_BYTEFIELD,
                    physical_type=DataType.A_BYTEFIELD),
        }

        # data object properties
        dops = {
            "certificateClient":
                DataObjectProperty(
                    odx_id=OdxLinkId("BV.dummy_DL.DOP.certificateClient", doc_frags),
                    short_name="certificateClient",
                    diag_coded_type=diagcodedtypes["certificateClient"],
                    physical_type=PhysicalType(base_data_type=DataType.A_BYTEFIELD),
                    compu_method=compumethods["bytes_passthrough"],
                ),
        }

        # Request
        request = Request(
            odx_id=OdxLinkId("BV.dummy_DL.RQ.sendCertificate", doc_frags),
            short_name="sendCertificate",
            parameters=NamedItemList([
                CodedConstParameter(
                    short_name="SID",
                    semantic="SERVICE-ID",
                    diag_coded_type=diagcodedtypes["uint8"],
                    byte_position=0,
                    coded_value_raw=str(uds.UDSSID.Authentication.value),
                ),
                ValueParameter(
                    short_name="certificateClient",
                    description=Description.from_string("The certificate to verify."),
                    byte_position=1,
                    # This DOP references the above parameter lengthOfCertificateClient for the bit length.
                    dop_ref=OdxLinkRef.from_id(dops["certificateClient"].odx_id),
                ),
            ]),
        )

        # Dummy diag layer to resolve references from request parameters to DOPs
        ecu_variant_raw = EcuVariantRaw(
            variant_type=DiagLayerType.ECU_VARIANT,
            odx_id=OdxLinkId("BV.dummy_DL", doc_frags),
            short_name="dummy_DL",
            diag_data_dictionary_spec=DiagDataDictionarySpec(
                data_object_props=NamedItemList(dops.values())),
            requests=NamedItemList([request]),
        )
        ecu_variant = EcuVariant(diag_layer_raw=ecu_variant_raw)
        odxlinks = OdxLinkDatabase()
        odxlinks.update(ecu_variant._build_odxlinks())
        db = Database()
        ecu_variant._resolve_odxlinks(odxlinks)
        ecu_variant._finalize_init(db, odxlinks)

        # Test decoding.
        coded_request = bytes([
            0x29,  # SID for Authentication
            0x03,  # Byte length of the certificate
            0x12,  # A very short
            0x34,  # certificate
            0x56,  # of three bytes
        ])
        self.assertEqual(
            request.decode(coded_request),
            {
                "SID": 0x29,
                "certificateClient": bytes([0x12, 0x34, 0x56])
            },
        )

        # Test encoding.
        self.assertEqual(
            request.encode(certificateClient=0x123456.to_bytes(3, "big")),
            0x2903123456.to_bytes(5, "big"),
        )


class TestStandardLengthType(unittest.TestCase):

    def test_decode_standard_length_type_uint(self) -> None:
        dct = StandardLengthType(
            base_data_type=DataType.A_UINT32,
            bit_length=5,
        )
        state = DecodeState(bytes([0x1, 0x72, 0x3]), cursor_byte_position=1, cursor_bit_position=1)
        internal_value = dct.decode_from_pdu(state)
        self.assertEqual(internal_value, 25)
        self.assertEqual(state.cursor_byte_position, 2)

    def test_decode_standard_length_type_uint_byteorder(self) -> None:
        dct = StandardLengthType(
            base_data_type=DataType.A_UINT32,
            bit_length=16,
            is_highlow_byte_order_raw=False,
        )
        state = DecodeState(bytes([0x1, 0x2, 0x3]), cursor_byte_position=1)
        internal_value = dct.decode_from_pdu(state)
        self.assertEqual(internal_value, 0x0302)
        self.assertEqual(state.cursor_byte_position, 3)

    def test_decode_standard_length_type_bytes(self) -> None:
        dct = StandardLengthType(
            base_data_type=DataType.A_BYTEFIELD,
            bit_length=16,
        )
        state = DecodeState(bytes([0x12, 0x34, 0x56, 0x78]), cursor_byte_position=1)
        internal_value = dct.decode_from_pdu(state)
        self.assertEqual(internal_value, bytes([0x34, 0x56]))
        self.assertEqual(state.cursor_byte_position, 3)


class TestParamLengthInfoType(unittest.TestCase):

    def test_decode_param_info_length_type_uint(self) -> None:
        length_key_id = OdxLinkId("param.length_key", doc_frags)
        length_key_ref = OdxLinkRef.from_id(length_key_id)
        length_key = LengthKeyParameter(
            odx_id=length_key_id,
            short_name="length_key",
            dop_ref=OdxLinkRef("DOP.uint8", doc_frags),
            byte_position=1,
        )
        dct = ParamLengthInfoType(
            base_data_type=DataType.A_UINT32,
            length_key_ref=length_key_ref,
        )
        odxlinks = OdxLinkDatabase()
        odxlinks.update({length_key_id: length_key})
        dct._resolve_odxlinks(odxlinks)
        state = DecodeState(
            coded_message=bytes([0x10, 0x12, 0x34, 0x56]),
            length_keys={length_key.short_name: 16},
            cursor_byte_position=1,
        )
        internal_value = dct.decode_from_pdu(state)
        self.assertEqual(internal_value, 0x1234)
        self.assertEqual(state.cursor_byte_position, 3)

    def test_encode_param_info_length_type_uint(self) -> None:
        length_key_id = OdxLinkId("param.length_key", doc_frags)
        length_key = LengthKeyParameter(
            odx_id=length_key_id,
            short_name="length_key",
            dop_ref=OdxLinkRef("DOP.uint8", doc_frags),
            byte_position=1,
        )
        dct = ParamLengthInfoType(
            base_data_type=DataType.A_UINT32,
            length_key_ref=OdxLinkRef.from_id(length_key_id),
        )
        odxlinks = OdxLinkDatabase()
        odxlinks.update({length_key_id: length_key})
        dct._resolve_odxlinks(odxlinks)
        state = EncodeState(coded_message=bytearray([0xcc]), cursor_byte_position=2)
        dct.encode_into_pdu(0x12345, state)
        self.assertEqual(state.coded_message.hex(), "cc00012345")
        self.assertEqual(state.length_keys.get("length_key"), 24)

    def test_end_to_end(self) -> None:
        # diag coded types
        diagcodedtypes = {
            "uint8":
                StandardLengthType(
                    base_data_type=DataType.A_UINT32,
                    bit_length=8,
                ),
            "length_key_id_to_lengthOfCertificateClient":
                ParamLengthInfoType(
                    base_data_type=DataType.A_UINT32,
                    length_key_ref=OdxLinkRef("param.dummy_length_key", doc_frags),
                ),
        }

        # computation methods
        compumethods = {
            "uint_passthrough":
                IdenticalCompuMethod(
                    category=CompuCategory.IDENTICAL,
                    internal_type=DataType.A_UINT32,
                    physical_type=DataType.A_UINT32),
            "multiply_with_8":
                LinearCompuMethod(
                    category=CompuCategory.LINEAR,
                    compu_internal_to_phys=CompuInternalToPhys(compu_scales=[
                        CompuScale(
                            compu_rational_coeffs=CompuRationalCoeffs(
                                value_type=DataType.A_INT32,
                                numerators=[0, 8],
                                denominators=[1],
                            ),
                            domain_type=DataType.A_INT32,
                            range_type=DataType.A_INT32),
                    ]),
                    internal_type=DataType.A_UINT32,
                    physical_type=DataType.A_UINT32,
                ),
        }

        # data object properties
        dops = {
            "uint8_times_8":
                DataObjectProperty(
                    odx_id=OdxLinkId("BV.dummy_DL.DOP.uint8_times_8", doc_frags),
                    short_name="uint8_times_8",
                    diag_coded_type=diagcodedtypes["uint8"],
                    physical_type=PhysicalType(base_data_type=DataType.A_UINT32),
                    compu_method=compumethods["multiply_with_8"],
                ),
            "certificateClient":
                DataObjectProperty(
                    odx_id=OdxLinkId("BV.dummy_DL.DOP.certificateClient", doc_frags),
                    short_name="certificateClient",
                    diag_coded_type=diagcodedtypes["length_key_id_to_lengthOfCertificateClient"],
                    physical_type=PhysicalType(base_data_type=DataType.A_UINT32),
                    compu_method=compumethods["uint_passthrough"],
                ),
        }

        # Request using LengthKeyParameter and ParamLengthInfoType
        request = Request(
            odx_id=OdxLinkId("BV.dummy_DL.RQ.sendCertificate", doc_frags),
            short_name="sendCertificate",
            parameters=NamedItemList([
                CodedConstParameter(
                    short_name="SID",
                    semantic="SERVICE-ID",
                    diag_coded_type=diagcodedtypes["uint8"],
                    byte_position=0,
                    coded_value_raw=str(uds.UDSSID.Authentication.value),
                ),
                LengthKeyParameter(
                    short_name="lengthOfCertificateClient",
                    description=Description.from_string("Length parameter for certificateClient."),
                    # LengthKeyParams have an ID to be referenced by a ParamLengthInfoType (which is a diag coded type)
                    odx_id=OdxLinkId("param.dummy_length_key", doc_frags),
                    byte_position=1,
                    # The DOP multiplies the coded value by 8, since the length key ref expects the number of bits.
                    dop_ref=OdxLinkRef.from_id(dops["uint8_times_8"].odx_id),
                ),
                ValueParameter(
                    short_name="certificateClient",
                    description=Description.from_string("The certificate to verify."),
                    byte_position=2,
                    # This DOP references the above parameter lengthOfCertificateClient for the bit length.
                    dop_ref=OdxLinkRef.from_id(dops["certificateClient"].odx_id),
                ),
            ]),
        )

        # Dummy diag layer to resolve references from request parameters to DOPs
        ecu_variant_raw = EcuVariantRaw(
            variant_type=DiagLayerType.BASE_VARIANT,
            odx_id=OdxLinkId("BV.dummy_DL", doc_frags),
            short_name="dummy_DL",
            diag_data_dictionary_spec=DiagDataDictionarySpec(
                data_object_props=NamedItemList(dops.values())),
            requests=NamedItemList([request]),
        )
        ecu_variant = EcuVariant(diag_layer_raw=ecu_variant_raw)
        odxlinks = OdxLinkDatabase()
        odxlinks.update(ecu_variant._build_odxlinks())
        db = Database()
        ecu_variant._resolve_odxlinks(odxlinks)
        ecu_variant._finalize_init(db, odxlinks)

        # Test decoding.
        coded_request = bytes([
            0x29,  # SID for Authentication
            0x03,  # Byte length of the certificate
            0x12,  # A very short
            0x34,  # certificate
            0x56,  # of three bytes
        ])

        self.assertEqual(
            request.decode(coded_request),
            {
                "SID": 0x29,
                "lengthOfCertificateClient": 24,
                "certificateClient": 0x123456
            },
        )

        # explicit defined length key
        encoded = request.encode(lengthOfCertificateClient=24, certificateClient=0x123456)
        self.assertEqual(encoded, coded_request)

        # Automatic bit length calculation
        encoded = request.encode(certificateClient=0x123456)
        self.assertEqual(encoded, coded_request)


class TestMinMaxLengthType(unittest.TestCase):

    def test_decode_min_max_length_type_bytes(self) -> None:
        dct = MinMaxLengthType(
            base_data_type=DataType.A_BYTEFIELD,
            min_length=1,
            max_length=4,
            termination=Termination.HEX_FF,
        )
        state = DecodeState(bytes([0x12, 0xFF, 0x34, 0x56, 0xFF]), cursor_byte_position=1)
        internal_value = dct.decode_from_pdu(state)
        self.assertEqual(internal_value, bytes([0xFF, 0x34, 0x56]))
        self.assertEqual(state.cursor_byte_position, 5)

    def test_decode_min_max_length_type_too_short_pdu(self) -> None:
        """If the PDU ends before min length is reached, an error must be raised."""
        dct = MinMaxLengthType(
            base_data_type=DataType.A_BYTEFIELD,
            min_length=2,
            max_length=4,
            termination=Termination.HEX_FF,
        )
        state = DecodeState(bytes([0x12, 0xFF]), cursor_byte_position=1)
        self.assertRaises(DecodeError, dct.decode_from_pdu, state)

    def test_decode_min_max_length_type_end_of_pdu(self) -> None:
        """If the PDU ends before max length is reached, the extracted value ends at the end of the PDU."""
        for termination in [Termination.END_OF_PDU, Termination.HEX_FF, Termination.ZERO]:
            dct = MinMaxLengthType(
                base_data_type=DataType.A_BYTEFIELD,
                min_length=2,
                max_length=5,
                termination=termination,
            )
            state = DecodeState(bytes([0x12, 0x34, 0x56, 0x78, 0x9A]), cursor_byte_position=1)
            internal_value = dct.decode_from_pdu(state)
            self.assertEqual(internal_value, bytes([0x34, 0x56, 0x78, 0x9A]))
            self.assertEqual(state.cursor_byte_position, 5)

    def test_decode_min_max_length_type_max_length(self) -> None:
        """If the max length is smaller than the end of PDU, the extracted value ends after max length."""
        for termination in [Termination.END_OF_PDU, Termination.HEX_FF, Termination.ZERO]:
            dct = MinMaxLengthType(
                base_data_type=DataType.A_BYTEFIELD,
                min_length=2,
                max_length=3,
                termination=termination,
            )
            state = DecodeState(bytes([0x12, 0x34, 0x56, 0x78, 0x9A]), cursor_byte_position=1)
            internal_value = dct.decode_from_pdu(state)
            self.assertEqual(internal_value, bytes([0x34, 0x56, 0x78]))
            self.assertEqual(state.cursor_byte_position, 4)

    def test_encode_min_max_length_type_hex_ff(self) -> None:
        dct = MinMaxLengthType(
            base_data_type=DataType.A_BYTEFIELD,
            min_length=1,
            max_length=4,
            termination=Termination.HEX_FF,
        )
        state = EncodeState(is_end_of_pdu=False)
        dct.encode_into_pdu(bytes([0x34, 0x56]), state)
        self.assertEqual(state.coded_message, bytes([0x34, 0x56, 0xFF]))

    def test_encode_min_max_length_type_zero(self) -> None:
        dct = MinMaxLengthType(
            base_data_type=DataType.A_UTF8STRING,
            min_length=2,
            max_length=4,
            termination=Termination.ZERO,
        )
        state = EncodeState(is_end_of_pdu=False)
        dct.encode_into_pdu("Hi", state)
        self.assertEqual(state.coded_message, bytes([0x48, 0x69, 0x0]))

    def test_encode_min_max_length_type_end_of_pdu(self) -> None:
        """If the parameter is at the end of the PDU, no termination char is added."""
        for termination in [Termination.END_OF_PDU, Termination.HEX_FF, Termination.ZERO]:
            dct = MinMaxLengthType(
                base_data_type=DataType.A_BYTEFIELD,
                min_length=2,
                max_length=5,
                termination=termination,
            )
            state = EncodeState(coded_message=bytearray(), is_end_of_pdu=True)
            dct.encode_into_pdu(bytes([0x34, 0x56, 0x78, 0x9A]), state)
            self.assertEqual(state.coded_message.hex(), "3456789a")

            if termination == Termination.END_OF_PDU:
                state = EncodeState(coded_message=bytearray(), is_end_of_pdu=False)
                self.assertRaises(OdxError, dct.encode_into_pdu, bytes([0x34, 0x56, 0x78, 0x9A]),
                                  state)
            else:
                state = EncodeState(coded_message=bytearray(), is_end_of_pdu=False)
                dct.encode_into_pdu(bytes([0x34, 0x56, 0x78, 0x9A]), state)
                self.assertTrue(state.coded_message.hex().startswith("3456789a"))

    def test_encode_min_max_length_type_min_length(self) -> None:
        """If the internal value is smaller than min length, an EncodeError must be raised."""
        for termination in [Termination.END_OF_PDU, Termination.HEX_FF, Termination.ZERO]:
            dct = MinMaxLengthType(
                base_data_type=DataType.A_BYTEFIELD,
                min_length=2,
                max_length=3,
                termination=termination,
            )
            state = EncodeState(is_end_of_pdu=True)
            dct.encode_into_pdu(bytes([0x34, 0x56]), state)
            self.assertTrue(state.coded_message.hex().startswith("3456"))
            self.assertRaises(
                EncodeError,
                dct.encode_into_pdu,
                bytes([0x34]),
                state,
            )

    def test_encode_min_max_length_type_max_length(self) -> None:
        """If the internal value is larger than max length, an EncodeError must be raised."""
        for termination in [Termination.END_OF_PDU, Termination.HEX_FF, Termination.ZERO]:
            dct = MinMaxLengthType(
                base_data_type=DataType.A_BYTEFIELD,
                min_length=2,
                max_length=3,
                termination=termination,
            )
            state = EncodeState(is_end_of_pdu=True)
            dct.encode_into_pdu(bytes([0x34, 0x56, 0x78]), state)
            self.assertEqual(state.coded_message, bytes([0x34, 0x56, 0x78]))
            self.assertRaises(
                EncodeError,
                dct.encode_into_pdu,
                bytes([0x34, 0x56, 0x78, 0x9A]),
                state,
            )

    def test_end_to_end(self) -> None:
        # diag coded types
        diagcodedtypes = {
            "uint8":
                StandardLengthType(
                    base_data_type=DataType.A_UINT32,
                    bit_length=8,
                ),
            "certificateClient":
                MinMaxLengthType(
                    base_data_type=DataType.A_BYTEFIELD,
                    min_length=2,
                    max_length=10,
                    termination=Termination.ZERO,
                ),
        }

        # computation methods
        compumethods = {
            "uint_passthrough":
                IdenticalCompuMethod(
                    category=CompuCategory.IDENTICAL,
                    internal_type=DataType.A_UINT32,
                    physical_type=DataType.A_UINT32),
            "bytes_passthrough":
                IdenticalCompuMethod(
                    category=CompuCategory.IDENTICAL,
                    internal_type=DataType.A_BYTEFIELD,
                    physical_type=DataType.A_BYTEFIELD),
        }

        # data object properties
        dops = {
            "certificateClient":
                DataObjectProperty(
                    odx_id=OdxLinkId("BV.dummy_DL.DOP.certificateClient", doc_frags),
                    short_name="certificateClient",
                    diag_coded_type=diagcodedtypes["certificateClient"],
                    physical_type=PhysicalType(base_data_type=DataType.A_BYTEFIELD),
                    compu_method=compumethods["bytes_passthrough"],
                ),
        }

        # Request
        request = Request(
            odx_id=OdxLinkId("BV.dummy_DL.RQ.sendCertificate", doc_frags),
            short_name="sendCertificate",
            parameters=NamedItemList([
                CodedConstParameter(
                    short_name="SID",
                    semantic="SERVICE-ID",
                    diag_coded_type=diagcodedtypes["uint8"],
                    byte_position=0,
                    coded_value_raw=str(uds.UDSSID.Authentication.value),
                ),
                ValueParameter(
                    short_name="certificateClient",
                    description=Description.from_string("The certificate to verify."),
                    byte_position=1,
                    # This DOP references the above parameter lengthOfCertificateClient for the bit length.
                    dop_ref=OdxLinkRef.from_id(dops["certificateClient"].odx_id),
                ),
                CodedConstParameter(
                    short_name="dummy",
                    diag_coded_type=diagcodedtypes["uint8"],
                    coded_value_raw=str(0x99),
                ),
            ]),
        )

        # Dummy diag layer to resolve references from request parameters to DOPs
        ecu_variant_raw = EcuVariantRaw(
            variant_type=DiagLayerType.BASE_VARIANT,
            odx_id=OdxLinkId("BV.dummy_DL", doc_frags),
            short_name="dummy_DL",
            diag_data_dictionary_spec=DiagDataDictionarySpec(
                data_object_props=NamedItemList(dops.values())),
            requests=NamedItemList([request]),
        )
        ecu_variant = EcuVariant(diag_layer_raw=ecu_variant_raw)
        odxlinks = OdxLinkDatabase()
        odxlinks.update(ecu_variant._build_odxlinks())
        db = Database()
        ecu_variant._resolve_odxlinks(odxlinks)
        ecu_variant._finalize_init(db, odxlinks)

        # Test decoding.
        coded_request = bytes([
            0x29,  # SID for Authentication
            0x12,  # A very short
            0x34,  # certificate
            0x56,  # of three bytes
            0x00,  # end of min-max length
            0x99,
        ])
        self.assertEqual(
            request.decode(coded_request),
            {
                "SID": 0x29,
                "certificateClient": bytes([0x12, 0x34, 0x56]),
                "dummy": 0x99
            },
        )

        # Test encoding.
        self.assertEqual(
            request.encode(certificateClient=0x123456.to_bytes(3, "big")), coded_request)

    def test_read_odx(self) -> None:
        expected = MinMaxLengthType(
            base_data_type=DataType.A_ASCIISTRING,
            base_type_encoding=Encoding.ISO_8859_1,
            min_length=8,
            max_length=16,
            termination=Termination.ZERO,
        )

        # diag-coded-type requires xsi namespace
        diagcodedtype_odx = f"""
        <ODX xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
            <DIAG-CODED-TYPE BASE-TYPE-ENCODING="{expected.base_type_encoding and expected.base_type_encoding.value}" BASE-DATA-TYPE="{expected.base_data_type.value}" TERMINATION="{expected.termination.value}" xsi:type="MIN-MAX-LENGTH-TYPE">
                <MIN-LENGTH>{expected.min_length}</MIN-LENGTH>
                <MAX-LENGTH>{expected.max_length}</MAX-LENGTH>
            </DIAG-CODED-TYPE>
        </ODX>
        """

        odx_element = ElementTree.fromstring(diagcodedtype_odx)
        diag_coded_type_element = odxrequire(odx_element.find("DIAG-CODED-TYPE"))

        actual = create_any_diag_coded_type_from_et(diag_coded_type_element,
                                                    OdxDocContext(Version("2.2.0"), doc_frags))

        self.assertIsInstance(actual, MinMaxLengthType)
        assert isinstance(actual, MinMaxLengthType)
        assert isinstance(expected, MinMaxLengthType)
        self.assertEqual(actual.base_data_type, expected.base_data_type)
        self.assertEqual(actual.base_type_encoding, expected.base_type_encoding)
        self.assertEqual(actual.min_length, expected.min_length)
        self.assertEqual(actual.max_length, expected.max_length)
        self.assertEqual(actual.termination, expected.termination)
        self.assertEqual(actual.is_highlow_byte_order, expected.is_highlow_byte_order)


if __name__ == "__main__":
    unittest.main()



# SPDX-License-Identifier: MIT
from dataclasses import dataclass, field
from xml.etree import ElementTree

from typing_extensions import override

from .exceptions import odxassert
from .matchingbasevariantparameter import MatchingBaseVariantParameter
from .matchingparameter import MatchingParameter
from .odxdoccontext import OdxDocContext
from .variantpattern import VariantPattern


@dataclass(kw_only=True)
class EcuVariantPattern(VariantPattern):
    """ECU variant patterns are variant patterns used to identify the
    concrete variant of an ECU.
    """
    matching_parameters: list[MatchingParameter] = field(default_factory=list)

    @override
    def get_matching_parameters(self
                               ) -> list[MatchingParameter] | list[MatchingBaseVariantParameter]:
        return self.matching_parameters

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "EcuVariantPattern":

        matching_parameters = [
            MatchingParameter.from_et(mp_el, context)
            for mp_el in et_element.iterfind("MATCHING-PARAMETERS/"
                                             "MATCHING-PARAMETER")
        ]

        odxassert(len(matching_parameters) > 0)  # required by ISO 22901-1 Figure 141
        return EcuVariantPattern(matching_parameters=matching_parameters)



# SPDX-License-Identifier: MIT
import typing
from dataclasses import dataclass
from typing import TYPE_CHECKING, runtime_checkable
from xml.etree import ElementTree

from .element import IdentifiableElement, NamedElement
from .nameditemlist import NamedItemList
from .odxdoccontext import OdxDocContext
from .utils import dataclass_fields_asdict

if TYPE_CHECKING:
    pass


@runtime_checkable
class HasVariableGroups(typing.Protocol):

    @property
    def variable_groups(self) -> NamedItemList["VariableGroup"]:
        ...


@dataclass(kw_only=True)
class VariableGroup(IdentifiableElement):

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "VariableGroup":
        kwargs = dataclass_fields_asdict(NamedElement.from_et(et_element, context))

        return VariableGroup(**kwargs)



# SPDX-License-Identifier: MIT
from dataclasses import dataclass

from .dopbase import DopBase


@dataclass(kw_only=True)
class ComplexDop(DopBase):
    """Base class for all complex data object properties.

    Complex DOPs are e.g., structures, fields and environment datas.
    """
    pass



# SPDX-License-Identifier: MIT
from dataclasses import dataclass
from typing import Any, Literal, cast
from xml.etree import ElementTree

from .decodestate import DecodeState
from .encodestate import EncodeState
from .encoding import Encoding
from .exceptions import odxassert, odxraise, odxrequire
from .odxdoccontext import OdxDocContext
from .odxlink import OdxLinkDatabase, OdxLinkId
from .odxtypes import AtomicOdxType, DataType, odxstr_to_bool
from .snrefcontext import SnRefContext

# Allowed diag-coded types
DctType = Literal[
    "LEADING-LENGTH-INFO-TYPE",
    "MIN-MAX-LENGTH-TYPE",
    "PARAM-LENGTH-INFO-TYPE",
    "STANDARD-LENGTH-TYPE",
]


@dataclass(kw_only=True)
class DiagCodedType:
    base_type_encoding: Encoding | None = None
    base_data_type: DataType

    is_highlow_byte_order_raw: bool | None = None

    @property
    def dct_type(self) -> DctType:
        odxraise(f"Class {type(self).__name__} does not override required method "
                 f"dct_type()", NotImplementedError)
        return cast(DctType, None)

    @property
    def is_highlow_byte_order(self) -> bool:
        return self.is_highlow_byte_order_raw in [None, True]

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "DiagCodedType":
        base_type_encoding = None
        if (base_type_encoding_str := et_element.get("BASE-TYPE-ENCODING")) is not None:
            try:
                base_type_encoding = Encoding(base_type_encoding_str)
            except ValueError:
                odxraise(f"Encountered unknown BASE-TYPE-ENCODING '{base_type_encoding_str}'")

        base_data_type_str = odxrequire(et_element.get("BASE-DATA-TYPE"))
        try:
            base_data_type = DataType(base_data_type_str)
        except ValueError:
            odxraise(f"Unknown base data type {base_data_type_str}")
            base_data_type = cast(DataType, None)

        is_highlow_byte_order_raw = odxstr_to_bool(et_element.get("IS-HIGHLOW-BYTE-ORDER"))

        return DiagCodedType(
            base_type_encoding=base_type_encoding,
            base_data_type=base_data_type,
            is_highlow_byte_order_raw=is_highlow_byte_order_raw)

    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:  # noqa: B027
        return {}

    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:  # noqa: B027
        """Recursively resolve any odxlinks references"""
        pass

    def _resolve_snrefs(self, context: SnRefContext) -> None:  # noqa: B027
        """Recursively resolve any short-name references"""
        pass

    def get_static_bit_length(self) -> int | None:
        return None

    def _minimal_byte_length_of(self, internal_value: bytes | str) -> int:
        """Helper method to get the minimal byte length.
        (needed for LeadingLength- and MinMaxLengthType)
        """
        byte_length: int = -1
        # A_BYTEFIELD, A_ASCIISTRING, A_UNICODE2STRING, A_UTF8STRING
        if self.base_data_type == DataType.A_BYTEFIELD:
            byte_length = len(internal_value)
        elif self.base_data_type in [DataType.A_ASCIISTRING, DataType.A_UTF8STRING]:
            if not isinstance(internal_value, str):
                odxraise()

            # TODO: Handle different encodings
            byte_length = len(bytes(internal_value, "utf-8"))
        elif self.base_data_type == DataType.A_UNICODE2STRING:
            if not isinstance(internal_value, str):
                odxraise()

            byte_length = len(bytes(internal_value, "utf-16-le"))
            odxassert(
                byte_length % 2 == 0, f"The bit length of A_UNICODE2STRING must"
                f" be a multiple of 16 but is {8*byte_length}")

        return byte_length

    def encode_into_pdu(self, internal_value: AtomicOdxType, encode_state: EncodeState) -> None:
        """Encode the internal value.

        Parameters
        ----------
        internal_value : python type corresponding to self.base_data_type
            the value to be encoded
        bit_position : int

        length_keys : Dict[OdxLinkId, int]
            mapping from ID (of the length key) to bit length
            (only needed for ParamLengthInfoType)
        """
        raise NotImplementedError(
            f".encode_into_pdu() is not implemented by the class {type(self).__name__}")

    def decode_from_pdu(self, decode_state: DecodeState) -> AtomicOdxType:
        """Decode the parameter value from the coded message.

        Parameters
        ----------
        decode_state : DecodeState
            The decoding state

        Returns
        -------
        str or int or bytes or dict
            the decoded parameter value
        int
            the next byte position after the extracted parameter
        """
        raise NotImplementedError(
            f".decode_from_pdu() is not implemented by the class {type(self).__name__}")



# SPDX-License-Identifier: MIT
from dataclasses import dataclass
from xml.etree import ElementTree

from typing_extensions import override

from ..decodestate import DecodeState
from ..encodestate import EncodeState
from ..exceptions import EncodeError, odxraise, odxrequire
from ..odxdoccontext import OdxDocContext
from ..odxtypes import DataType, ParameterValue
from ..utils import dataclass_fields_asdict
from .parameter import Parameter, ParameterType


@dataclass(kw_only=True)
class MatchingRequestParameter(Parameter):
    request_byte_position: int
    byte_length: int

    @property
    @override
    def parameter_type(self) -> ParameterType:
        return "MATCHING-REQUEST-PARAM"

    @property
    @override
    def is_required(self) -> bool:
        return False

    @property
    @override
    def is_settable(self) -> bool:
        return False

    @staticmethod
    @override
    def from_et(et_element: ElementTree.Element,
                context: OdxDocContext) -> "MatchingRequestParameter":

        kwargs = dataclass_fields_asdict(Parameter.from_et(et_element, context))

        request_byte_position = int(odxrequire(et_element.findtext("REQUEST-BYTE-POS")))
        byte_length = int(odxrequire(et_element.findtext("BYTE-LENGTH")))

        return MatchingRequestParameter(
            request_byte_position=request_byte_position, byte_length=byte_length, **kwargs)

    @override
    def get_static_bit_length(self) -> int | None:
        return 8 * self.byte_length

    @override
    def _encode_positioned_into_pdu(self, physical_value: ParameterValue | None,
                                    encode_state: EncodeState) -> None:
        if encode_state.triggering_request is None:
            odxraise(
                f"Parameter '{self.short_name}' is of matching request type,"
                f" but no original request has been specified.", EncodeError)
            return

        rq_pos = self.request_byte_position
        rq_len = self.byte_length

        if len(encode_state.triggering_request) < rq_pos + rq_len:
            odxraise(
                f"Specified triggering request 0x{encode_state.triggering_request.hex()} "
                f"is not long enough to encode matching request parameter "
                f"'{self.short_name}': Have {len(encode_state.triggering_request)} "
                f"bytes, need at least {rq_pos + rq_len} bytes", EncodeError)
            return

        encode_state.emplace_bytes(encode_state.triggering_request[rq_pos:rq_pos + rq_len],
                                   self.short_name)

    @override
    def _decode_positioned_from_pdu(self, decode_state: DecodeState) -> ParameterValue:
        return decode_state.extract_atomic_value(
            bit_length=self.byte_length * 8,
            base_data_type=DataType.A_UINT32,
            base_type_encoding=None,
            is_highlow_byte_order=False)



# SPDX-License-Identifier: MIT
from dataclasses import dataclass, field
from typing import Any
from xml.etree import ElementTree

from ..comparaminstance import ComparamInstance
from ..odxdoccontext import OdxDocContext
from ..odxlink import OdxLinkDatabase, OdxLinkId
from ..snrefcontext import SnRefContext
from ..utils import dataclass_fields_asdict
from .diaglayerraw import DiagLayerRaw


@dataclass(kw_only=True)
class HierarchyElementRaw(DiagLayerRaw):
    """This is the base class for diagnostic layers that may be involved in value inheritance

    This class represents the data present in the XML, not the "logical" view.
    """

    comparam_refs: list[ComparamInstance] = field(default_factory=list)

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "HierarchyElementRaw":
        # objects contained by diagnostic layers exibit an additional
        # document fragment for the diag layer, so we use the document
        # fragments of the odx id of the diag layer for IDs of
        # contained objects.
        dlr = DiagLayerRaw.from_et(et_element, context)
        kwargs = dataclass_fields_asdict(dlr)

        comparam_refs = [
            ComparamInstance.from_et(el, context)
            for el in et_element.iterfind("COMPARAM-REFS/COMPARAM-REF")
        ]

        return HierarchyElementRaw(comparam_refs=comparam_refs, **kwargs)

    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        result = super()._build_odxlinks()

        for comparam_ref in self.comparam_refs:
            result.update(comparam_ref._build_odxlinks())

        return result

    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        super()._resolve_odxlinks(odxlinks)

        for comparam_ref in self.comparam_refs:
            comparam_ref._resolve_odxlinks(odxlinks)

    def _resolve_snrefs(self, context: SnRefContext) -> None:
        super()._resolve_snrefs(context)

        for comparam_ref in self.comparam_refs:
            comparam_ref._resolve_snrefs(context)



# SPDX-License-Identifier: MIT

from .compuconst import CompuConst

# make CompuInverseValue an alias for CompuConst. The XSD defines two
# separate but identical groups for this (why?)...
CompuInverseValue = CompuConst



# SPDX-License-Identifier: MIT
import argparse

from ..database import Database
from ..diagservice import DiagService
from ..odxtypes import ParameterValue
from ..singleecujob import SingleEcuJob
from . import _parser_utils
from ._parser_utils import SubparsersList
from ._print_utils import print_diagnostic_service

# name of the tool
_odxtools_tool_name_ = "find"


def get_display_value(v: ParameterValue) -> str:
    if isinstance(v, bytes):
        return v.hex(" ")
    elif isinstance(v, int):
        return f"{v} (0x{v:x})"
    else:
        return str(v)


def print_summary(odxdb: Database,
                  service_names: list[str],
                  ecu_variants: list[str] | None = None,
                  allow_unknown_bit_lengths: bool = False,
                  print_params: bool = False) -> None:
    ecu_names = ecu_variants if ecu_variants else [ecu.short_name for ecu in odxdb.ecus]
    service_db: dict[str, DiagService] = {}
    service_ecus: dict[str, list[str]] = {}
    for ecu_name in ecu_names:
        ecu = odxdb.ecus[ecu_name]
        if not ecu:
            print(f"The ecu variant '{ecu_name}' could not be found!")
            continue

        if service_names:
            for service_name_search in service_names:
                for service in ecu.services:
                    if service_name_search.lower() in service.short_name.lower():
                        ecu_names = service_ecus.get(service.short_name, [])
                        ecu_names.append(ecu_name)
                        service_ecus[service.short_name] = ecu_names
                        service_db[service.short_name] = service

    for service_name, ecu_names in service_ecus.items():
        service = service_db[service_name]
        display_names = ", ".join(ecu_names)
        filler = str.ljust("", len(display_names), "=")
        print(f"\n{filler}")
        print(f"{', '.join(ecu_names)}")
        print(f"{filler}\n\n")
        if isinstance(service, DiagService):
            print_diagnostic_service(
                service,
                print_params=print_params,
                allow_unknown_bit_lengths=allow_unknown_bit_lengths,
                print_pre_condition_states=True,
                print_state_transitions=True,
                print_audiences=True)
        elif isinstance(service, SingleEcuJob):
            print(f"SingleEcuJob: {service.odx_id}")
        else:
            print(f"Unknown service: {service}")


def add_subparser(subparsers: SubparsersList) -> None:
    parser = subparsers.add_parser(
        "find",
        description="\n".join([
            "Find & print services by name",
            "",
            "Examples:",
            "  For displaying the services associated with the partial name 'Reset' without details:",
            '    odxtools find ./path/to/database.pdx -s "Reset" --no-details',
            "  For more information use:",
            "    odxtools find -h",
        ]),
        help="Find & display services by their name",
        formatter_class=argparse.RawTextHelpFormatter,
    )
    _parser_utils.add_pdx_argument(parser)

    parser.add_argument(
        "-v",
        "--variants",
        nargs=1,
        metavar="VARIANT",
        required=False,
        help="Specifies which ecu variants should be included.",
        default="all",
    )

    parser.add_argument(
        "-s",
        "--service-names",
        nargs="*",
        default=None,
        metavar="SERVICES",
        required=True,
        help="Print a list of diagnostic services partially matching given service names",
    )

    parser.add_argument(
        "-nd",
        "--no-details",
        action="store_true",
        required=False,
        help="Don't show all service details",
    )

    parser.add_argument(
        "-ro",
        "--relaxed-output",
        action="store_true",
        required=False,
        help="Relax output formatting rules (allow unknown bitlengths for ascii representation)",
    )


def run(args: argparse.Namespace) -> None:
    odxdb = _parser_utils.load_file(args)
    variants = args.variants

    print_summary(
        odxdb,
        ecu_variants=None if variants == "all" else variants,
        service_names=args.service_names,
        print_params=not args.no_details,
        allow_unknown_bit_lengths=args.relaxed_output)



# SPDX-License-Identifier: MIT
from dataclasses import dataclass
from typing import Any
from xml.etree import ElementTree

from .dopbase import DopBase
from .element import NamedElement
from .exceptions import odxrequire
from .odxdoccontext import OdxDocContext
from .odxlink import OdxLinkDatabase, OdxLinkId, OdxLinkRef
from .snrefcontext import SnRefContext
from .utils import dataclass_fields_asdict


@dataclass(kw_only=True)
class NegOutputParam(NamedElement):
    dop_base_ref: OdxLinkRef

    @property
    def dop(self) -> DopBase:
        """The data object property describing this parameter."""
        return self._dop

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "NegOutputParam":

        kwargs = dataclass_fields_asdict(NamedElement.from_et(et_element, context))
        dop_base_ref = odxrequire(OdxLinkRef.from_et(et_element.find("DOP-BASE-REF"), context))

        return NegOutputParam(dop_base_ref=dop_base_ref, **kwargs)

    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        return {}

    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        self._dop = odxlinks.resolve(self.dop_base_ref, DopBase)

    def _resolve_snrefs(self, context: SnRefContext) -> None:
        pass



# SPDX-License-Identifier: MIT
from dataclasses import dataclass
from typing import Any
from xml.etree import ElementTree

from typing_extensions import override

from ..dataobjectproperty import DataObjectProperty
from ..encodestate import EncodeState
from ..exceptions import EncodeError, odxraise, odxrequire
from ..odxdoccontext import OdxDocContext
from ..odxlink import OdxLinkDatabase, OdxLinkId
from ..odxtypes import AtomicOdxType, ParameterValue
from ..snrefcontext import SnRefContext
from ..utils import dataclass_fields_asdict
from .parameter import ParameterType
from .parameterwithdop import ParameterWithDOP


@dataclass(kw_only=True)
class ValueParameter(ParameterWithDOP):
    physical_default_value_raw: str | None = None

    @property
    @override
    def parameter_type(self) -> ParameterType:
        return "VALUE"

    @property
    def physical_default_value(self) -> AtomicOdxType | None:
        return self._physical_default_value

    @property
    @override
    def is_required(self) -> bool:
        return self._physical_default_value is None

    @property
    @override
    def is_settable(self) -> bool:
        return True

    @staticmethod
    @override
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "ValueParameter":

        kwargs = dataclass_fields_asdict(ParameterWithDOP.from_et(et_element, context))

        physical_default_value_raw = et_element.findtext("PHYSICAL-DEFAULT-VALUE")

        return ValueParameter(physical_default_value_raw=physical_default_value_raw, **kwargs)

    def __post_init__(self) -> None:
        self._physical_default_value: AtomicOdxType | None = None

    @override
    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        return super()._build_odxlinks()

    @override
    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        super()._resolve_odxlinks(odxlinks)

    @override
    def _resolve_snrefs(self, context: SnRefContext) -> None:
        super()._resolve_snrefs(context)

        if self.physical_default_value_raw is not None:
            dop = odxrequire(self.dop)
            if not isinstance(dop, DataObjectProperty):
                odxraise("Value parameters can only define a physical default "
                         "value if they use a simple DOP")
            base_data_type = dop.physical_type.base_data_type
            self._physical_default_value = base_data_type.from_string(
                self.physical_default_value_raw)

    @override
    def _encode_positioned_into_pdu(self, physical_value: ParameterValue | None,
                                    encode_state: EncodeState) -> None:

        if physical_value is None:
            physical_value = self._physical_default_value
        if physical_value is None:
            odxraise(
                f"A value for parameter '{self.short_name}' must be specified"
                f" because the parameter does not exhibit a default.", EncodeError)

        self.dop.encode_into_pdu(physical_value, encode_state=encode_state)



# SPDX-License-Identifier: MIT
from collections.abc import Iterable
from copy import deepcopy
from dataclasses import dataclass
from typing import Any, cast
from xml.etree import ElementTree

from typing_extensions import override

from ..diagvariable import DiagVariable
from ..exceptions import odxassert
from ..nameditemlist import NamedItemList
from ..odxdoccontext import OdxDocContext
from ..odxlink import OdxLinkDatabase, OdxLinkRef
from ..parentref import ParentRef
from ..variablegroup import VariableGroup
from .diaglayer import DiagLayer
from .functionalgroupraw import FunctionalGroupRaw
from .hierarchyelement import HierarchyElement


@dataclass(kw_only=True)
class FunctionalGroup(HierarchyElement):
    """This is a diagnostic layer for functionality shared between multiple ECU variants
    """

    @property
    def functional_group_raw(self) -> FunctionalGroupRaw:
        return cast(FunctionalGroupRaw, self.diag_layer_raw)

    @property
    def diag_variables_raw(self) -> list[DiagVariable | OdxLinkRef]:
        return self.functional_group_raw.diag_variables_raw

    @property
    def diag_variables(self) -> NamedItemList[DiagVariable]:
        return self._diag_variables

    @property
    def variable_groups(self) -> NamedItemList[VariableGroup]:
        return self._variable_groups

    @property
    def parent_refs(self) -> list[ParentRef]:
        return self.functional_group_raw.parent_refs

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "FunctionalGroup":
        functional_group_raw = FunctionalGroupRaw.from_et(et_element, context)

        return FunctionalGroup(diag_layer_raw=functional_group_raw)

    def __post_init__(self) -> None:
        super().__post_init__()

        odxassert(
            isinstance(self.diag_layer_raw, FunctionalGroupRaw),
            "The raw diagnostic layer passed to FunctionalGroup "
            "must be a FunctionalGroupRaw")

    @override
    def _compute_value_inheritance(self, odxlinks: OdxLinkDatabase) -> None:
        super()._compute_value_inheritance(odxlinks)

        self._diag_variables = NamedItemList(self._compute_available_diag_variables(odxlinks))
        self._variable_groups = NamedItemList(self._compute_available_variable_groups(odxlinks))

    def _compute_available_diag_variables(self,
                                          odxlinks: OdxLinkDatabase) -> Iterable[DiagVariable]:

        def get_local_objects_fn(dl: DiagLayer) -> Iterable[DiagVariable]:
            if not hasattr(dl.diag_layer_raw, "diag_variables"):
                return []

            return dl.diag_layer_raw.diag_variables  # type: ignore[no-any-return]

        def not_inherited_fn(parent_ref: ParentRef) -> list[str]:
            return parent_ref.not_inherited_variables

        return self._compute_available_objects(get_local_objects_fn, not_inherited_fn)

    def _compute_available_variable_groups(self,
                                           odxlinks: OdxLinkDatabase) -> Iterable[VariableGroup]:

        def get_local_objects_fn(dl: DiagLayer) -> Iterable[VariableGroup]:
            if not hasattr(dl.diag_layer_raw, "variable_groups"):
                return []

            return dl.diag_layer_raw.variable_groups  # type: ignore[no-any-return]

        def not_inherited_fn(parent_ref: ParentRef) -> list[str]:
            return []

        return self._compute_available_objects(get_local_objects_fn, not_inherited_fn)

    def __deepcopy__(self, memo: dict[int, Any]) -> Any:
        """Create a deep copy of the functional group layer

        Note that the copied diagnostic layer is not fully
        initialized, so `_finalize_init()` should to be called on it
        before it can be used normally.
        """

        result = super().__deepcopy__(memo)

        # note that the self.functional_group_raw object is *not* copied at
        # this place because the attribute points to the same object
        # as self.diag_layer_raw.
        result.functional_group_raw = deepcopy(self.functional_group_raw, memo)

        return result



# SPDX-License-Identifier: MIT
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any
from xml.etree import ElementTree

from .nameditemlist import NamedItemList
from .odxcategory import OdxCategory
from .odxdoccontext import OdxDocContext
from .odxlink import DocType, OdxLinkDatabase, OdxLinkId
from .protstack import ProtStack
from .snrefcontext import SnRefContext
from .utils import dataclass_fields_asdict

if TYPE_CHECKING:
    from .database import Database


@dataclass(kw_only=True)
class ComparamSpec(OdxCategory):

    prot_stacks: NamedItemList[ProtStack] = field(default_factory=NamedItemList)

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "ComparamSpec":

        base_obj = OdxCategory.category_from_et(et_element, context, doc_type=DocType.COMPARAM_SPEC)
        kwargs = dataclass_fields_asdict(base_obj)

        prot_stacks = NamedItemList([
            ProtStack.from_et(dl_element, context)
            for dl_element in et_element.iterfind("PROT-STACKS/PROT-STACK")
        ])

        return ComparamSpec(prot_stacks=prot_stacks, **kwargs)

    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        odxlinks = super()._build_odxlinks()

        for ps in self.prot_stacks:
            odxlinks.update(ps._build_odxlinks())

        return odxlinks

    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        super()._resolve_odxlinks(odxlinks)

        for ps in self.prot_stacks:
            ps._resolve_odxlinks(odxlinks)

    def _finalize_init(self, database: "Database", odxlinks: OdxLinkDatabase) -> None:
        super()._finalize_init(database, odxlinks)

    def _resolve_snrefs(self, context: SnRefContext) -> None:
        super()._resolve_snrefs(context)

        for ps in self.prot_stacks:
            ps._resolve_snrefs(context)



# SPDX-License-Identifier: MIT
from dataclasses import dataclass
from xml.etree import ElementTree

from ..odxtypes import AtomicOdxType, DataType


@dataclass(kw_only=True)
class CompuConst:
    v: str | None = None
    vt: str | None = None

    data_type: DataType

    @property
    def value(self) -> AtomicOdxType | None:
        return self._value

    @staticmethod
    def compuvalue_from_et(et_element: ElementTree.Element, *, data_type: DataType) -> "CompuConst":

        v = et_element.findtext("V")
        vt = et_element.findtext("VT")

        return CompuConst(v=v, vt=vt, data_type=data_type)

    def __post_init__(self) -> None:
        self._value: AtomicOdxType | None = self.vt
        if self.v is not None:
            self._value = self.data_type.from_string(self.v)



# SPDX-License-Identifier: MIT
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, cast
from xml.etree import ElementTree

from typing_extensions import override

from .decodestate import DecodeState
from .diagcodedtype import DctType, DiagCodedType
from .encodestate import EncodeState
from .exceptions import EncodeError, odxraise, odxrequire
from .odxdoccontext import OdxDocContext
from .odxlink import OdxLinkDatabase, OdxLinkId, OdxLinkRef
from .odxtypes import AtomicOdxType, DataType
from .snrefcontext import SnRefContext
from .utils import dataclass_fields_asdict

if TYPE_CHECKING:
    from .parameters.lengthkeyparameter import LengthKeyParameter


@dataclass(kw_only=True)
class ParamLengthInfoType(DiagCodedType):
    length_key_ref: OdxLinkRef

    @property
    def dct_type(self) -> DctType:
        return "PARAM-LENGTH-INFO-TYPE"

    @property
    def length_key(self) -> "LengthKeyParameter":
        return self._length_key

    @staticmethod
    @override
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "ParamLengthInfoType":
        kwargs = dataclass_fields_asdict(DiagCodedType.from_et(et_element, context))

        length_key_ref = odxrequire(OdxLinkRef.from_et(et_element.find("LENGTH-KEY-REF"), context))

        return ParamLengthInfoType(length_key_ref=length_key_ref, **kwargs)

    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        return super()._build_odxlinks()

    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        """Recursively resolve any odxlinks references"""
        super()._resolve_odxlinks(odxlinks)

        if TYPE_CHECKING:
            self._length_key = odxlinks.resolve(self.length_key_ref, LengthKeyParameter)
        else:
            self._length_key = odxlinks.resolve(self.length_key_ref)

    def _resolve_snrefs(self, context: SnRefContext) -> None:
        """Recursively resolve any short-name references"""
        super()._resolve_snrefs(context)

    @override
    def encode_into_pdu(self, internal_value: AtomicOdxType, encode_state: EncodeState) -> None:
        bit_length = encode_state.length_keys.get(self.length_key.short_name)

        if bit_length is None:
            # the length key is implicit, i.e., we need to set the
            # value for the length key in the encode_state based on
            # the value passed here.
            if self.base_data_type in [
                    DataType.A_BYTEFIELD,
                    DataType.A_ASCIISTRING,
                    DataType.A_UTF8STRING,
            ]:
                bit_length = 8 * len(cast(str, internal_value))
            elif self.base_data_type in [DataType.A_UNICODE2STRING]:
                bit_length = 16 * len(cast(str, internal_value))
            elif self.base_data_type in [DataType.A_INT32, DataType.A_UINT32]:
                if not isinstance(internal_value, int):
                    odxraise()
                bit_length = int(internal_value).bit_length()
                if self.base_data_type == DataType.A_INT32:
                    bit_length += 1
                # Round up
                bit_length = ((bit_length + 7) // 8) * 8
            elif self.base_data_type == DataType.A_FLOAT32:
                bit_length = 32
            elif self.base_data_type == DataType.A_FLOAT64:
                bit_length = 64
            else:
                odxraise(
                    f"Cannot determine size of an object of type "
                    f"{self.base_data_type.value}", EncodeError)
                return

            encode_state.length_keys[self.length_key.short_name] = bit_length

        encode_state.emplace_atomic_value(
            internal_value=internal_value,
            used_mask=None,
            bit_length=bit_length,
            base_data_type=self.base_data_type,
            base_type_encoding=self.base_type_encoding,
            is_highlow_byte_order=self.is_highlow_byte_order,
        )

    def decode_from_pdu(self, decode_state: DecodeState) -> AtomicOdxType:
        # First, we need to find a length key with matching ID.
        if self.length_key.short_name not in decode_state.length_keys:
            odxraise(f"Unspecified mandatory length key parameter "
                     f"{self.length_key.short_name}")
            decode_state.cursor_bit_position = 0
            return cast(None, AtomicOdxType)

        bit_length = decode_state.length_keys[self.length_key.short_name]
        if not isinstance(bit_length, int):
            odxraise(f"The bit length must be an integer, is {type(bit_length)}")
            bit_length = 0

        # Extract the internal value and return.
        return decode_state.extract_atomic_value(
            bit_length=bit_length,
            base_data_type=self.base_data_type,
            base_type_encoding=self.base_type_encoding,
            is_highlow_byte_order=self.is_highlow_byte_order,
        )



# SPDX-License-Identifier: MIT
from dataclasses import dataclass, field
from typing import Any
from xml.etree import ElementTree

from .admindata import AdminData
from .nameditemlist import NamedItemList
from .odxdoccontext import OdxDocContext
from .odxlink import OdxLinkDatabase, OdxLinkId
from .physicaldimension import PhysicalDimension
from .snrefcontext import SnRefContext
from .specialdatagroup import SpecialDataGroup
from .unit import Unit
from .unitgroup import UnitGroup


@dataclass(kw_only=True)
class UnitSpec:
    """
    A unit spec encapsulates three lists:

    * unit groups
    * units
    * physical_dimensions

    The following odx elements are not internalized: ADMIN-DATA, SDGS
    """

    admin_data: AdminData | None = None
    unit_groups: NamedItemList[UnitGroup] = field(default_factory=NamedItemList)
    units: NamedItemList[Unit] = field(default_factory=NamedItemList)
    physical_dimensions: NamedItemList[PhysicalDimension] = field(default_factory=NamedItemList)
    sdgs: list[SpecialDataGroup] = field(default_factory=list)

    def __post_init__(self) -> None:
        self.unit_groups = NamedItemList(self.unit_groups)
        self.units = NamedItemList(self.units)
        self.physical_dimensions = NamedItemList(self.physical_dimensions)

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "UnitSpec":

        admin_data = AdminData.from_et(et_element.find("ADMIN-DATA"), context)
        unit_groups = NamedItemList([
            UnitGroup.from_et(el, context) for el in et_element.iterfind("UNIT-GROUPS/UNIT-GROUP")
        ])
        units = NamedItemList(
            [Unit.from_et(el, context) for el in et_element.iterfind("UNITS/UNIT")])
        physical_dimensions = NamedItemList([
            PhysicalDimension.from_et(el, context)
            for el in et_element.iterfind("PHYSICAL-DIMENSIONS/PHYSICAL-DIMENSION")
        ])
        sdgs = [SpecialDataGroup.from_et(sdge, context) for sdge in et_element.iterfind("SDGS/SDG")]

        return UnitSpec(
            admin_data=admin_data,
            unit_groups=unit_groups,
            units=units,
            physical_dimensions=physical_dimensions,
            sdgs=sdgs)

    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        odxlinks: dict[OdxLinkId, Any] = {}
        for unit in self.units:
            odxlinks.update(unit._build_odxlinks())
        for dim in self.physical_dimensions:
            odxlinks.update(dim._build_odxlinks())
        for sdg in self.sdgs:
            odxlinks.update(sdg._build_odxlinks())

        return odxlinks

    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        for unit in self.units:
            unit._resolve_odxlinks(odxlinks)
        for group in self.unit_groups:
            group._resolve_odxlinks(odxlinks)
        for sdg in self.sdgs:
            sdg._resolve_odxlinks(odxlinks)

    def _resolve_snrefs(self, context: SnRefContext) -> None:
        for unit in self.units:
            unit._resolve_snrefs(context)
        for group in self.unit_groups:
            group._resolve_snrefs(context)
        for sdg in self.sdgs:
            sdg._resolve_snrefs(context)



# SPDX-License-Identifier: MIT
from dataclasses import dataclass
from typing import cast
from xml.etree import ElementTree

from ..exceptions import DecodeError, EncodeError, odxassert, odxraise
from ..odxdoccontext import OdxDocContext
from ..odxtypes import AtomicOdxType, DataType
from ..utils import dataclass_fields_asdict
from .compucategory import CompuCategory
from .compumethod import CompuMethod
from .ratfuncsegment import RatFuncSegment


@dataclass(kw_only=True)
class ScaleRatFuncCompuMethod(CompuMethod):
    """A compu method using a piecewise rational function

    For details, refer to ASAM specification MCD-2 D (ODX), section 7.3.6.6.5.
    """

    @property
    def int_to_phys_segments(self) -> list[RatFuncSegment]:
        return self._int_to_phys_segments

    @property
    def phys_to_int_segments(self) -> list[RatFuncSegment] | None:
        return self._phys_to_int_segments

    @staticmethod
    def compu_method_from_et(et_element: ElementTree.Element, context: OdxDocContext, *,
                             internal_type: DataType,
                             physical_type: DataType) -> "ScaleRatFuncCompuMethod":
        cm = CompuMethod.compu_method_from_et(
            et_element, context, internal_type=internal_type, physical_type=physical_type)
        kwargs = dataclass_fields_asdict(cm)

        return ScaleRatFuncCompuMethod(**kwargs)

    def __post_init__(self) -> None:
        odxassert(self.category == CompuCategory.SCALE_RAT_FUNC,
                  "ScaleRatFuncCompuMethod must exhibit SCALE-RAT-FUNC category")

        odxassert(self.physical_type in [
            DataType.A_FLOAT32,
            DataType.A_FLOAT64,
            DataType.A_INT32,
            DataType.A_UINT32,
        ])
        odxassert(self.internal_type in [
            DataType.A_FLOAT32,
            DataType.A_FLOAT64,
            DataType.A_INT32,
            DataType.A_UINT32,
        ])

        if self.compu_internal_to_phys is None:
            odxraise("RAT-FUNC compu methods require COMPU-INTERNAL-TO-PHYS")
            return

        int_to_phys_scales = self.compu_internal_to_phys.compu_scales
        if len(int_to_phys_scales) < 1:
            odxraise("RAT-FUNC compu methods expect at least one compu scale within "
                     "COMPU-INTERNAL-TO-PHYS")
            return

        self._int_to_phys_segments = [
            RatFuncSegment.from_compu_scale(scale, value_type=self.physical_type)
            for scale in int_to_phys_scales
        ]

        self._phys_to_int_segments = None
        if self.compu_phys_to_internal is not None:
            phys_to_int_scales = self.compu_phys_to_internal.compu_scales
            if len(phys_to_int_scales) < 1:
                odxraise("SCALE-RAT-FUNC compu methods expect at least one compu scale within "
                         "COMPU-PHYS-TO-INTERNAL")
                return

            self._phys_to_int_segments = [
                RatFuncSegment.from_compu_scale(scale, value_type=self.internal_type)
                for scale in phys_to_int_scales
            ]

    def convert_internal_to_physical(self, internal_value: AtomicOdxType) -> AtomicOdxType:
        for seg in self._int_to_phys_segments:
            if seg.applies(internal_value):
                return seg.convert(internal_value)

        odxraise(f"Internal value {internal_value!r} be decoded using this compumethod",
                 DecodeError)
        return cast(AtomicOdxType, None)

    def convert_physical_to_internal(self, physical_value: AtomicOdxType) -> AtomicOdxType:
        if self._phys_to_int_segments is None:
            odxraise(f"Physical values cannot be encoded using this compumethod", EncodeError)
            return cast(AtomicOdxType, None)

        for seg in self._phys_to_int_segments:
            if seg.applies(physical_value):
                return seg.convert(physical_value)

        odxraise(f"Physical values {physical_value!r} be decoded using this compumethod",
                 EncodeError)
        return cast(AtomicOdxType, None)

    def is_valid_internal_value(self, internal_value: AtomicOdxType) -> bool:
        return any(seg.applies(internal_value) for seg in self._int_to_phys_segments)

    def is_valid_physical_value(self, physical_value: AtomicOdxType) -> bool:
        if self._phys_to_int_segments is None:
            return False

        return any(seg.applies(physical_value) for seg in self._phys_to_int_segments)



# SPDX-License-Identifier: MIT
import warnings
from collections.abc import Iterable
from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Optional, TypeVar, overload
from xml.etree import ElementTree

from .exceptions import OdxWarning, odxassert, odxraise, odxrequire
from .nameditemlist import OdxNamed, TNamed
from .odxdoccontext import OdxDocContext


class DocType(Enum):
    FLASH = "FLASH"
    CONTAINER = "CONTAINER"
    LAYER = "LAYER"
    MULTIPLE_ECU_JOB_SPEC = "MULTIPLE-ECU-JOB-SPEC"
    COMPARAM_SPEC = "COMPARAM-SPEC"
    VEHICLE_INFO_SPEC = "VEHICLE-INFO-SPEC"
    COMPARAM_SUBSET = "COMPARAM-SUBSET"
    ECU_CONFIG = "ECU-CONFIG"
    FUNCTION_DICTIONARY_SPEC = "FUNCTION-DICTIONARY-SPEC"


@dataclass(frozen=True)
class OdxDocFragment:
    doc_name: str
    doc_type: DocType


@dataclass(frozen=True)
class OdxLinkId:
    """The identifier of an ODX object.

    A full ODX ID comprises the document fragment where the object
    is located (i.e., the short name object beneath the root object of the
    currently read XMl file) plus a locally unique identifier for the object.

    OdxLinkIds are hashable.
    """

    #: An identifier that is (hopefully) locally unique within the
    #: relevant document fragment
    local_id: str

    #: The name and type of the document fragment to which the
    #: `local_id` is relative to
    doc_fragments: list[OdxDocFragment] = field(default_factory=list)

    def __hash__(self) -> int:
        # we do not hash about the document fragment here, because
        # document fragments are handled using separate sub-databases,
        # i.e. the same OdxId object can be put into all of them.
        return hash(self.local_id)

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, OdxLinkId):
            return False

        # if the local ID is different, the whole id is different
        if self.local_id != other.local_id:
            return False

        # the document fragments must be identical for the IDs to be identical
        return self.doc_fragments == other.doc_fragments

    def __str__(self) -> str:
        return f"OdxLinkId('{self.local_id}')"

    @staticmethod
    def from_et(et: ElementTree.Element, context: OdxDocContext) -> Optional["OdxLinkId"]:
        """Construct an OdxLinkId for a given XML node (ElementTree object).

        Returns None if the given XML node does not exhibit an ID.
        """

        local_id = et.attrib.get("ID")
        if local_id is None:
            return None

        return OdxLinkId(local_id, context.doc_fragments)


@dataclass
class OdxLinkRef:
    """A reference to an ODX object.

    OdxLinkRefs are similar to OdxLinkIds, with the difference that the document
    of the referred-to object does not need to be specified. (In this case, it
    means "use the document where the referring object is located in")
    """

    #: The local identifier of the object which is referred to
    ref_id: str

    #: The document fragments to which the `ref_id` refers to (in reverse order)
    ref_docs: list[OdxDocFragment] = field(default_factory=list)

    # TODO: this is difficult because OdxLinkRef is derived from and
    # we do not want having to specify it mandatorily
    #revision: Optional[str] = None

    @overload
    @staticmethod
    def from_et(et: None, context: OdxDocContext) -> None:
        ...

    @overload
    @staticmethod
    def from_et(et: ElementTree.Element, context: OdxDocContext) -> "OdxLinkRef":
        ...

    @staticmethod
    def from_et(et: ElementTree.Element | None, context: OdxDocContext) -> Optional["OdxLinkRef"]:
        """Construct an OdxLinkRef for a given XML node (ElementTree object).

        Returns None if the given XML node does not represent a reference.
        """

        if et is None:
            return None

        id_ref = et.attrib.get("ID-REF")
        if id_ref is None:
            odxraise(f"Tag {et.tag} is not a ODXLINK reference")
            return None

        docref = et.attrib.get("DOCREF")
        doctype_attr = et.attrib.get("DOCTYPE")

        odxassert((docref is not None and doctype_attr is not None) or
                  (docref is None and doctype_attr is None),
                  "DOCREF and DOCTYPE must both either be specified or omitted")

        doctype = None
        if doctype_attr is not None:
            try:
                doctype = DocType(doctype_attr)
            except ValueError:
                odxraise(f"Encountered unknown document type '{doctype_attr}'")

        # if the target document fragment is specified by the
        # reference, use it, else use the document fragment containing
        # the reference.
        if docref is not None:
            doc_frags = [OdxDocFragment(docref, odxrequire(doctype))]
        else:
            doc_frags = context.doc_fragments

        return OdxLinkRef(id_ref, doc_frags)

    @staticmethod
    def from_id(odxid: OdxLinkId) -> "OdxLinkRef":
        """Construct an OdxLinkRef for a given OdxLinkId."""
        return OdxLinkRef(odxid.local_id, odxid.doc_fragments)

    def __str__(self) -> str:
        return f"OdxLinkRef('{self.ref_id}')"


T = TypeVar("T")


class OdxLinkDatabase:
    """
    A database holding all objects which exhibit OdxLinkIds

    This can resolve ODXLINK references.
    """

    def __init__(self) -> None:
        self._db: dict[OdxDocFragment, dict[str, Any]] = {}

    @overload
    def resolve(self, ref: OdxLinkRef, expected_type: None = None) -> Any:
        ...

    @overload
    def resolve(self, ref: OdxLinkRef, expected_type: type[T]) -> T:
        ...

    def resolve(self, ref: OdxLinkRef, expected_type: Any | None = None) -> Any:
        """
        Resolve a reference to an object

        If the database does not contain any object which is referred to, a
        KeyError exception is raised.
        """
        for ref_frag in reversed(ref.ref_docs):
            doc_frag_db = self._db.get(ref_frag)
            if doc_frag_db is None:
                # No object featured by the database uses the document
                # fragment mentioned by the reference. This should not
                # happen for correct ODX databases...
                warnings.warn(
                    f"Warning: Unknown document fragment {ref_frag} "
                    f"when resolving reference {ref}",
                    OdxWarning,
                    stacklevel=1,
                )
                continue

            # locate an object exhibiting with the referenced local ID
            # in the ID database for the document fragment
            if (obj := doc_frag_db.get(ref.ref_id)) is not None:
                if expected_type is not None:
                    odxassert(isinstance(obj, expected_type))

                return obj

        odxraise(
            f"ODXLINK reference {ref} could not be resolved for any "
            f"of the document fragments {ref.ref_docs}", KeyError)
        return None

    @overload
    def resolve_lenient(self, ref: OdxLinkRef, expected_type: None = None) -> Any:
        ...

    @overload
    def resolve_lenient(self, ref: OdxLinkRef, expected_type: type[T]) -> T | None:
        ...

    def resolve_lenient(self, ref: OdxLinkRef, expected_type: Any | None = None) -> Any | None:
        """
        Resolve a reference to an object

        If the database does not contain any object which is referred to, None
        is returned.
        """

        for ref_frag in reversed(ref.ref_docs):
            doc_frag_db = self._db.get(ref_frag)
            if doc_frag_db is None:
                # No object featured by the database uses the document
                # fragment mentioned by the reference. This should not
                # happen for correct databases...
                warnings.warn(
                    f"Warning: Unknown document fragment {ref_frag} "
                    f"when resolving reference {ref}",
                    OdxWarning,
                    stacklevel=1,
                )
                continue

            if (obj := doc_frag_db.get(ref.ref_id)) is not None:
                if expected_type is not None:
                    odxassert(isinstance(obj, expected_type))

                return obj

        return None

    def update(self, new_entries: dict[OdxLinkId, Any], *, overwrite: bool = True) -> None:
        """
        Add a bunch of new objects to the ODXLINK database.

        The argument needs to be an OdxLinkId -> object dictionary.
        """

        # put all new objects into the databases for the document
        # fragments which it specifies
        for odx_id, obj in new_entries.items():
            for doc_frag in odx_id.doc_fragments:
                if doc_frag not in self._db:
                    self._db[doc_frag] = {}

                if overwrite:
                    self._db[doc_frag][odx_id.local_id] = obj
                else:
                    self._db[doc_frag].setdefault(odx_id.local_id, obj)


@overload
def resolve_snref(target_short_name: str,
                  items: Iterable[OdxNamed],
                  expected_type: None = None,
                  *,
                  lenient: None = None) -> Any:
    """Resolve a short name reference given a sequence of candidate objects"""
    ...


@overload
def resolve_snref(target_short_name: str,
                  items: Iterable[OdxNamed],
                  expected_type: type[TNamed],
                  *,
                  lenient: None = None) -> TNamed:
    ...


@overload
def resolve_snref(target_short_name: str,
                  items: Iterable[OdxNamed],
                  expected_type: type[TNamed],
                  *,
                  lenient: bool = True) -> TNamed | None:
    ...


def resolve_snref(target_short_name: str,
                  items: Iterable[OdxNamed],
                  expected_type: Any = None,
                  lenient: bool | None = None) -> Any:
    candidates = [x for x in items if x.short_name == target_short_name]

    if not candidates:
        if not lenient:
            odxraise(f"Cannot resolve short name reference to '{target_short_name}'")
        return None
    elif len(candidates) > 1:
        odxraise(f"Cannot uniquely resolve short name reference to '{target_short_name}'")
    elif expected_type is not None and not isinstance(candidates[0], expected_type):
        odxraise(f"Reference '{target_short_name}' points to a {type(candidates[0]).__name__}"
                 f"object while expecting {expected_type.__name__}")

    return candidates[0]



# SPDX-License-Identifier: MIT
from dataclasses import dataclass
from typing import Any
from xml.etree import ElementTree

from .element import IdentifiableElement
from .exceptions import odxrequire
from .externalaccessmethod import ExternalAccessMethod
from .odxdoccontext import OdxDocContext
from .odxlink import OdxLinkDatabase, OdxLinkId, resolve_snref
from .snrefcontext import SnRefContext
from .state import State
from .utils import dataclass_fields_asdict


@dataclass(kw_only=True)
class StateTransition(IdentifiableElement):
    """
    Corresponds to STATE-TRANSITION.
    """
    source_snref: str
    target_snref: str
    external_access_method: ExternalAccessMethod | None = None

    @property
    def source_state(self) -> State:
        return self._source_state

    @property
    def target_state(self) -> State:
        return self._target_state

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "StateTransition":

        kwargs = dataclass_fields_asdict(IdentifiableElement.from_et(et_element, context))

        source_snref_elem = odxrequire(et_element.find("SOURCE-SNREF"))
        source_snref = odxrequire(source_snref_elem.attrib["SHORT-NAME"])

        target_snref_elem = odxrequire(et_element.find("TARGET-SNREF"))
        target_snref = odxrequire(target_snref_elem.attrib["SHORT-NAME"])

        external_access_method = None
        if (eam_elem := et_element.find("EXTERNAL-ACCESS-METHOD")) is not None:
            external_access_method = ExternalAccessMethod.from_et(eam_elem, context)
        return StateTransition(
            source_snref=source_snref,
            target_snref=target_snref,
            external_access_method=external_access_method,
            **kwargs)

    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        return {self.odx_id: self}

    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        pass

    def _resolve_snrefs(self, context: SnRefContext) -> None:
        states = odxrequire(context.state_chart).states
        self._source_state = resolve_snref(self.source_snref, states, State)
        self._target_state = resolve_snref(self.target_snref, states, State)



# SPDX-License-Identifier: MIT
from dataclasses import dataclass
from typing import Any
from xml.etree import ElementTree

from .companydata import CompanyData
from .exceptions import odxrequire
from .odxdoccontext import OdxDocContext
from .odxlink import OdxLinkDatabase, OdxLinkId, OdxLinkRef
from .snrefcontext import SnRefContext


@dataclass(kw_only=True)
class CompanyRevisionInfo:
    company_data_ref: OdxLinkRef
    revision_label: str | None = None
    state: str | None = None

    @property
    def company_data(self) -> CompanyData:
        return self._company_data

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "CompanyRevisionInfo":

        company_data_ref = odxrequire(
            OdxLinkRef.from_et(et_element.find("COMPANY-DATA-REF"), context))
        revision_label = et_element.findtext("REVISION-LABEL")
        state = et_element.findtext("STATE")

        return CompanyRevisionInfo(
            company_data_ref=company_data_ref, revision_label=revision_label, state=state)

    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        return {}

    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        self._company_data = odxlinks.resolve(self.company_data_ref, CompanyData)

    def _resolve_snrefs(self, context: SnRefContext) -> None:
        pass



# SPDX-License-Identifier: MIT
from dataclasses import dataclass
from xml.etree import ElementTree

from typing_extensions import override

from .decodestate import DecodeState
from .diagcodedtype import DctType, DiagCodedType
from .encodestate import EncodeState
from .exceptions import EncodeError, odxassert, odxraise, odxrequire
from .odxdoccontext import OdxDocContext
from .odxtypes import AtomicOdxType, DataType
from .utils import dataclass_fields_asdict


@dataclass(kw_only=True)
class LeadingLengthInfoType(DiagCodedType):
    #: bit length of the length specifier field
    #:
    #: this is then followed by the number of bytes specified by that
    #: field, i.e., this is NOT the length of the LeadingLengthInfoType
    #: object.
    bit_length: int

    @property
    def dct_type(self) -> DctType:
        return "LEADING-LENGTH-INFO-TYPE"

    @staticmethod
    @override
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "LeadingLengthInfoType":
        kwargs = dataclass_fields_asdict(DiagCodedType.from_et(et_element, context))

        bit_length = int(odxrequire(et_element.findtext("BIT-LENGTH")))

        return LeadingLengthInfoType(bit_length=bit_length, **kwargs)

    def __post_init__(self) -> None:
        odxassert(self.bit_length > 0,
                  "A Leading length info type with bit length == 0 does not make sense.")
        odxassert(
            self.base_data_type in [
                DataType.A_BYTEFIELD,
                DataType.A_ASCIISTRING,
                DataType.A_UNICODE2STRING,
                DataType.A_UTF8STRING,
            ],
            f"A leading length info type cannot have the base data type {self.base_data_type.name}."
        )

    @override
    def encode_into_pdu(self, internal_value: AtomicOdxType, encode_state: EncodeState) -> None:

        if not isinstance(internal_value, (str, bytes)):
            odxraise(
                f"LEADING-LENGTH-INFO types can only be used for strings and byte fields, "
                f"not {type(internal_value).__name__}", EncodeError)
            return

        byte_length = self._minimal_byte_length_of(internal_value)

        used_mask = None
        bit_pos = encode_state.cursor_bit_position
        if encode_state.cursor_bit_position != 0 or (bit_pos + self.bit_length) % 8 != 0:
            used_mask = (1 << self.bit_length) - 1
            used_mask <<= bit_pos

        encode_state.emplace_atomic_value(
            internal_value=byte_length,
            used_mask=None,
            bit_length=self.bit_length,
            base_data_type=DataType.A_UINT32,
            base_type_encoding=None,
            is_highlow_byte_order=self.is_highlow_byte_order,
        )

        encode_state.emplace_atomic_value(
            internal_value=internal_value,
            used_mask=None,
            bit_length=8 * byte_length,
            base_data_type=self.base_data_type,
            base_type_encoding=None,
            is_highlow_byte_order=self.is_highlow_byte_order,
        )

    @override
    def decode_from_pdu(self, decode_state: DecodeState) -> AtomicOdxType:

        # Extract length of the parameter value
        byte_length = decode_state.extract_atomic_value(
            bit_length=self.bit_length,
            base_data_type=DataType.A_UINT32,  # length is an integer
            base_type_encoding=None,
            is_highlow_byte_order=self.is_highlow_byte_order,
        )

        if not isinstance(byte_length, int):
            odxraise()

        # Extract actual value
        # TODO: The returned value is None if the byte_length is 0. Maybe change it
        #       to some default value like an empty bytearray() or 0?
        value = decode_state.extract_atomic_value(
            bit_length=8 * byte_length,
            base_data_type=self.base_data_type,
            base_type_encoding=None,
            is_highlow_byte_order=self.is_highlow_byte_order,
        )

        return value



# SPDX-License-Identifier: MIT
from collections.abc import Iterable, Iterator
from io import StringIO

from . import obd, uds
from .diagservice import DiagService
from .nameditemlist import NamedItemList
from .parameters.codedconstparameter import CodedConstParameter


class ServiceBinner:
    """Class to categorize a long list of services into the service
    groups defined by the UDS or OBD standards.

    This class is supposed to be used like this:

    db = odxtools.load_file("my_cool_diagnostics_db.pdx")
    ...
    service_groups = ServiceBinner(db.ecus.my_ecu.services)
    print(service_groups)
    """

    def __init__(self, services: Iterable[DiagService]):
        service_groups: dict[int | None, NamedItemList[DiagService]] = {}
        for service in services:
            SID = self.__extract_sid(service)

            if SID not in service_groups:
                service_groups[SID] = NamedItemList()

            service_groups[SID].append(service)

        self._service_groups = service_groups

    def __extract_sid(self, service: DiagService) -> int | None:
        # diagnostic services without requests are possible; just like
        # aircraft without wings...
        if service.request is None:
            return None

        prefix = 0  # prefix of constant parameters
        cursor = 0  # bit position of the next parameter
        for param in service.request.parameters:
            if not isinstance(param, CodedConstParameter):
                # we *need* at least the first byte of a request to be statically defined!
                return None

            param_len = param.get_static_bit_length()
            if param_len is None:
                return None

            if not isinstance(param.coded_value, int):
                return None

            prefix <<= param_len
            prefix |= param.coded_value & ((1 << param_len) - 1)
            cursor += param_len

            if cursor >= 8:
                # we have a prefix that is at least 8 bits
                # long. return its most significant byte.
                prefix >>= cursor - 8
                return prefix & 0xff

        return None

    def __str__(self) -> str:
        """Return an informative string about which of the diagnostic
        services are part of which UDS or OBD service group.

        """
        result = StringIO()
        for sid, service_list in self._service_groups.items():
            if isinstance(sid, int):
                sid_name = obd.sid_to_name(sid)
                if sid_name is not None:
                    print(f"OBD service group '{sid_name}' (0x{sid:x}):", file=result)
                elif (sid_name := uds.sid_to_name(sid)) is not None:
                    print(f"UDS service group '{sid_name}' (0x{sid:x}):", file=result)
                else:
                    print(f"Unknown service group 0x{sid:x}:", file=result)
            else:
                print(f"Non-standard services:", file=result)
            for service in service_list:
                print(f"  {service.short_name}", file=result)

        return result.getvalue()

    def __repr__(self) -> str:
        """Return an string representing the object
        """
        result = StringIO()
        result.write("[ ")
        result.write(", ".join([f"0x{x:x}" for x in self._service_groups if x is not None]))
        result.write(" ]")

        return result.getvalue()

    def __iter__(self) -> Iterator[int | None]:
        return iter(self._service_groups)

    def __getitem__(self, sid: int | None) -> NamedItemList[DiagService]:
        if sid is None:
            return self._service_groups.get(sid, NamedItemList())

        if sid < 0 or sid > 255:
            raise IndexError(f"SIDs must be in range 0x00 to 0xff")

        return self._service_groups.get(sid, NamedItemList())



# SPDX-License-Identifier: MIT
from dataclasses import dataclass
from typing import Optional, overload
from xml.etree import ElementTree

from .exceptions import odxraise, odxrequire
from .odxdoccontext import OdxDocContext
from .odxlink import OdxLinkRef
from .utils import dataclass_fields_asdict


@dataclass(kw_only=True)
class DynEndDopRef(OdxLinkRef):
    termination_value_raw: str

    @staticmethod
    @overload
    def from_et(et_element: None, context: OdxDocContext) -> None:
        ...

    @staticmethod
    @overload
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "DynEndDopRef":
        ...

    @staticmethod
    def from_et(et_element: ElementTree.Element | None,
                context: OdxDocContext) -> Optional["DynEndDopRef"]:

        if et_element is None:
            odxraise("Mandatory DYN-END-DOP-REF tag is missing")
            return None

        kwargs = dataclass_fields_asdict(OdxLinkRef.from_et(et_element, context))

        termination_value_raw = odxrequire(et_element.findtext("TERMINATION-VALUE"))

        return DynEndDopRef(termination_value_raw=termination_value_raw, **kwargs)



# SPDX-License-Identifier: MIT
from collections.abc import Generator
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

from .exceptions import odxraise
from .odxtypes import ParameterValueDict
from .state import State
from .statechart import StateChart

if TYPE_CHECKING:
    from .diaglayers.diaglayer import DiagLayer
    from .diagservice import DiagService


@dataclass(kw_only=True)
class StateMachine:
    """Objects of this class represent the runtime state of a state chart

    It is used to track which services can be executed at a given time.

    Usage example (using asynchronous communication routines):

    ```python
    ecu = db.ecu_variants.my_ecu_variant
    fsm = StateMachine(ecu, ecu.state_charts.my_state_chart)

    for raw_request in (executor := fsm.execute(ecu.services.my_service,
                                                param1="hello",
                                                param2=123)):
        # send raw request to the ECU (usually using ISO-TP or DoIP), and
        # receive the response
        await send_to_ecu(raw_request)
        while odxtools.is_response_pending(raw_response := await receive_from_ecu()):
            pass

        executor.send(raw_response)

    print(f"Request send: {fsm.succeeded}")
    ```

    Alternatively, more of the glue functionality can be handled by
    the calling code:

    ```python

    ecu = db.ecu_variants.my_ecu_variant
    fsm = StateMachine(ecu, ecu.state_charts.my_state_chart)

    service = ecu.services.my_service
    request_param_values = { "param1": "hello", "param2":123 }
    if any(pcs.applies(fsm, service.request.parameters, **request_param_values)
           for pcs in service.pre_condition_states):

        # send raw request to the ECU (usually using ISO-TP or DoIP), and
        # receive the response
        raw_req = service.request.encode(**request_param_values)
        await send_data_to_ecu(raw_req)
        while odxtools.is_response_pending(raw_resp := await receive_from_ecu()):
            pass

        done = False
        for decoded_resp_msg in ecu.decode_response(raw_resp, raw_req):
            for stransref in service.state_transition_refs:
                if stransref.execute(decoded_resp.parameters, decoded_resp_msg.param_dict):
                   done = True
                   break
            if done:
               break
    else:
        raise RuntimeException(f"Cannot execute request for service {service.short_name}")
    ```

    """

    succeeded: bool

    @property
    def diag_layer(self) -> "DiagLayer":
        return self._diag_layer

    @property
    def state_chart(self) -> StateChart:
        return self._state_chart

    @property
    def active_state(self) -> State:
        return self._active_state

    @active_state.setter
    def active_state(self, value: State) -> None:
        self._active_state = value

    def __init__(self, diag_layer: "DiagLayer", state_chart: StateChart) -> None:
        self.succeeded = True
        self._diag_layer = diag_layer
        self._state_chart = state_chart
        self._active_state = state_chart.start_state

    def execute(self, service: "DiagService", **service_params: Any
               ) -> Generator[bytes, bytes | bytearray | ParameterValueDict, None]:
        """Run a diagnostic service and update the state machine
        depending on the outcome.

        This simplifies error handling, en- and decoding etc.

        Usage example (using asynchronous communication routines):

        ```python
        ecu = db.ecu_variants.my_ecu_variant
        fsm = StateMachine(ecu, ecu.state_charts.my_state_chart)

        for raw_request in (executor := fsm.execute(ecu.services.my_service,
                                                    param1="hello",
                                                    param2=123)):
            # send raw request to the ECU (usually using ISO-TP or DoIP), and
            # receive the response
            await send_to_ecu(raw_request)
            while odxtools.is_response_pending(raw_response := await receive_from_ecu()):
                pass

            executor.send(raw_response)

        print(f"Request send: {fsm.succeeded}")
        ```

        """
        if service.request is None:
            odxraise("Services without requests are not allowed in this context")
            self.succeeded = False
            return

        # the service can be executed if any of the specified
        # precondition states is fulfilled. (TODO: correct?)
        if service.pre_condition_state_refs is not None:
            if not any(
                    x.applies(self, service.request.parameters, service_params)
                    for x in service.pre_condition_state_refs):
                # if all preconditions which are applicable are
                # invalid (i.e., they evaluate to False), we must not
                # execute the service.
                self.succeeded = False
                return

        raw_req = service.request.encode(**service_params)
        # ask the calling code to send the request to the ECU and
        # report back the reply
        raw_resp = yield raw_req

        decoded_req_params = service.request.decode(raw_req)
        for stransref in service.state_transition_refs:
            # check if the state transition applies for the
            # request. Note that the ODX specification is unclear
            # about which kind of parameters are relevant for
            # STATE-TRANSITION-REF
            if stransref.execute(self, service.request.parameters, decoded_req_params):
                self.succeeded = True
                yield b''
                return

        if raw_resp is None:
            raise RuntimeError("The calling code must send back a reply")
        elif isinstance(raw_resp, (bytes, bytearray)):
            for decoded_resp_msg in self.diag_layer.decode_response(raw_resp, raw_req):
                for stransref in service.state_transition_refs:
                    # we only execute the first applicable state
                    # transition: The spec seems to imply a
                    # deterministic state machine and chaining
                    # transistions most likely is not what the user
                    # expects. (The spec seems to be a bit loose on
                    # this front...)
                    if stransref.execute(self, decoded_resp_msg.coding_object.parameters,
                                         decoded_resp_msg.param_dict):
                        self.succeeded = True
                        yield b''
                        return
        else:
            resp_object = service.positive_responses[0]
            for stransref in service.state_transition_refs:
                if stransref.execute(self, resp_object.parameters, raw_resp):
                    self.succeeded = True
                    yield b''
                    return

        self.succeeded = True
        yield b''
        return



# SPDX-License-Identifier: MIT
import typing
from typing import runtime_checkable

from .codec import Codec
from .decodestate import DecodeState
from .encodestate import EncodeState
from .exceptions import EncodeError, odxraise
from .odxtypes import ParameterValue
from .parameters.codedconstparameter import CodedConstParameter
from .parameters.matchingrequestparameter import MatchingRequestParameter
from .parameters.parameter import Parameter
from .parameters.physicalconstantparameter import PhysicalConstantParameter


@runtime_checkable
class CompositeCodec(Codec, typing.Protocol):
    """Any object which can be en- or decoded to be transferred over
    the wire which is composed of multiple parameter implements this
    API.

    """

    @property
    def parameters(self) -> list[Parameter]:
        return []

    @property
    def required_parameters(self) -> list[Parameter]:
        return []

    @property
    def free_parameters(self) -> list[Parameter]:
        return []


# some helper functions useful for composite codec objects
def composite_codec_get_static_bit_length(codec: CompositeCodec) -> int | None:
    """Compute the length of a composite codec object in bits

    This is basically the sum of the lengths of all parameters. If the
    length of any parameter can only determined at runtime, `None` is
    returned.
    """

    cursor = 0
    byte_length = 0
    for param in codec.parameters:
        param_bit_length = param.get_static_bit_length()
        if param_bit_length is None:
            # We were not able to calculate a static bit length
            return None
        elif param.byte_position is not None:
            cursor = param.byte_position

        cursor += ((param.bit_position or 0) + param_bit_length + 7) // 8
        byte_length = max(byte_length, cursor)

    return byte_length * 8


def composite_codec_get_required_parameters(codec: CompositeCodec) -> list[Parameter]:
    """Return the list of parameters which are required to be
    specified for encoding the composite codec object

    I.e., all free parameters that do not exhibit a default value.
    """
    return [p for p in codec.parameters if p.is_required]


def composite_codec_get_free_parameters(codec: CompositeCodec) -> list[Parameter]:
    """Return the list of parameters which can be freely specified by
    the user when encoding the composite codec object

    This means all required parameters plus parameters that can be
    omitted because they specify a default.
    """
    return [p for p in codec.parameters if p.is_settable]


def composite_codec_get_coded_const_prefix(codec: CompositeCodec,
                                           request_prefix: bytes = b'') -> bytes:
    encode_state = EncodeState(coded_message=bytearray(), triggering_request=request_prefix)

    for param in codec.parameters:
        if (isinstance(param, MatchingRequestParameter) and param.request_byte_position < len(request_prefix)) or \
            isinstance(param, (CodedConstParameter, PhysicalConstantParameter)):
            param.encode_into_pdu(physical_value=None, encode_state=encode_state)
        else:
            break

    return encode_state.coded_message


def composite_codec_encode_into_pdu(codec: CompositeCodec, physical_value: ParameterValue | None,
                                    encode_state: EncodeState) -> None:
    from .parameters.lengthkeyparameter import LengthKeyParameter
    from .parameters.tablekeyparameter import TableKeyParameter

    if not isinstance(physical_value, dict):
        odxraise(
            f"Expected a dictionary for the values of {codec.short_name}, "
            f"got {type(physical_value).__name__}", EncodeError)
    elif encode_state.cursor_bit_position != 0:
        odxraise(
            f"Compositional codec objecs must be byte aligned, but "
            f"{codec.short_name} requested to be at bit position "
            f"{encode_state.cursor_bit_position}", EncodeError)
        encode_state.bit_position = 0

    orig_origin = encode_state.origin_byte_position
    encode_state.origin_byte_position = encode_state.cursor_byte_position

    orig_is_end_of_pdu = encode_state.is_end_of_pdu
    encode_state.is_end_of_pdu = False

    # ensure that no values for unknown parameters are specified.
    if not encode_state.allow_unknown_parameters:
        param_names = {param.short_name for param in codec.parameters}
        for param_value_name in physical_value:
            if param_value_name not in param_names:
                odxraise(f"Value for unknown parameter '{param_value_name}' specified "
                         f"for composite codec object {codec.short_name}")

    for param in codec.parameters:
        if id(param) == id(codec.parameters[-1]):
            # The last parameter of the composite codec object is at
            # the end of the PDU if the codec object itself is at the
            # end of the PDU.
            #
            # TODO: This assumes that the last parameter specified in
            # the ODX is located last in the PDU...
            encode_state.is_end_of_pdu = orig_is_end_of_pdu

        if isinstance(param, (LengthKeyParameter, TableKeyParameter)):
            # At this point, we encode a placeholder value for length-
            # and table keys, since these can be specified
            # implicitly (i.e., by means of parameters that use
            # these keys). To avoid getting an "overlapping
            # parameter" warning, we must encode a value of zero
            # into the PDU here and add the real value of the
            # parameter in a post-processing step.
            param.encode_placeholder_into_pdu(
                physical_value=physical_value.get(param.short_name), encode_state=encode_state)

            continue

        if param.is_required and param.short_name not in physical_value:
            odxraise(f"No value for required parameter {param.short_name} specified", EncodeError)

        param_phys_value = physical_value.get(param.short_name)
        param.encode_into_pdu(physical_value=param_phys_value, encode_state=encode_state)

        encode_state.journal.append((param, param_phys_value))

    encode_state.is_end_of_pdu = False

    # encode the length- and table keys. This cannot be done above
    # because we allow these to be defined implicitly (i.e. they
    # are defined by their respective users)
    for param in codec.parameters:
        if not isinstance(param, (LengthKeyParameter, TableKeyParameter)):
            # the current parameter is neither a length- nor a table key
            continue

        # Encode the value of the key parameter into the message
        param.encode_value_into_pdu(encode_state=encode_state)

    encode_state.origin_byte_position = orig_origin


def composite_codec_decode_from_pdu(codec: CompositeCodec,
                                    decode_state: DecodeState) -> ParameterValue:
    # move the origin since positions specified by sub-parameters of
    # composite codec objects are relative to the beginning of the
    # object.
    orig_origin = decode_state.origin_byte_position
    decode_state.origin_byte_position = decode_state.cursor_byte_position

    result = {}
    for param in codec.parameters:
        value = param.decode_from_pdu(decode_state)

        decode_state.journal.append((param, value))
        result[param.short_name] = value

    # decoding of the composite codec object finished. go back the
    # original origin.
    decode_state.origin_byte_position = orig_origin

    return result



# SPDX-License-Identifier: MIT
from dataclasses import dataclass, field
from typing import Any
from xml.etree import ElementTree

from ..odxdoccontext import OdxDocContext
from ..odxlink import OdxLinkDatabase, OdxLinkId
from ..odxtypes import DataType
from ..progcode import ProgCode
from ..snrefcontext import SnRefContext
from .compudefaultvalue import CompuDefaultValue
from .compuscale import CompuScale


@dataclass(kw_only=True)
class CompuInternalToPhys:
    compu_scales: list[CompuScale] = field(default_factory=list)
    prog_code: ProgCode | None = None
    compu_default_value: CompuDefaultValue | None = None

    @staticmethod
    def compu_internal_to_phys_from_et(et_element: ElementTree.Element, context: OdxDocContext, *,
                                       internal_type: DataType,
                                       physical_type: DataType) -> "CompuInternalToPhys":
        compu_scales = [
            CompuScale.compuscale_from_et(
                cse, context, domain_type=internal_type, range_type=physical_type)
            for cse in et_element.iterfind("COMPU-SCALES/COMPU-SCALE")
        ]

        prog_code = None
        if (pce := et_element.find("PROG-CODE")) is not None:
            prog_code = ProgCode.from_et(pce, context)

        compu_default_value = None
        if (cdve := et_element.find("COMPU-DEFAULT-VALUE")) is not None:
            compu_default_value = CompuDefaultValue.compuvalue_from_et(
                cdve, data_type=physical_type)

        return CompuInternalToPhys(
            compu_scales=compu_scales, prog_code=prog_code, compu_default_value=compu_default_value)

    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        result = {}

        if self.prog_code is not None:
            result.update(self.prog_code._build_odxlinks())

        return result

    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        if self.prog_code is not None:
            self.prog_code._resolve_odxlinks(odxlinks)

    def _resolve_snrefs(self, context: SnRefContext) -> None:
        if self.prog_code is not None:
            self.prog_code._resolve_snrefs(context)



# SPDX-License-Identifier: MIT
from dataclasses import dataclass
from xml.etree import ElementTree

from ..exceptions import odxassert
from ..odxdoccontext import OdxDocContext
from ..odxtypes import AtomicOdxType, DataType
from ..utils import dataclass_fields_asdict
from .compumethod import CompuMethod


@dataclass(kw_only=True)
class IdenticalCompuMethod(CompuMethod):
    """Identical compu methods just pass through the internal value.

    For details, refer to ASAM specification MCD-2 D (ODX), section 7.3.6.6.2.
    """

    @staticmethod
    def compu_method_from_et(et_element: ElementTree.Element, context: OdxDocContext, *,
                             internal_type: DataType,
                             physical_type: DataType) -> "IdenticalCompuMethod":
        cm = CompuMethod.compu_method_from_et(
            et_element, context, internal_type=internal_type, physical_type=physical_type)
        kwargs = dataclass_fields_asdict(cm)

        odxassert(
            internal_type == physical_type or
            (internal_type
             in [DataType.A_ASCIISTRING, DataType.A_UTF8STRING, DataType.A_UNICODE2STRING] and
             physical_type
             in [DataType.A_ASCIISTRING, DataType.A_UTF8STRING, DataType.A_UNICODE2STRING]),
            f"Internal type and physical type must be the same for compu methods of category "
            f"'{cm.category}' (internal type: '{internal_type.value}', physical type: "
            f"'{physical_type.value}')")

        return IdenticalCompuMethod(**kwargs)

    def convert_physical_to_internal(self, physical_value: AtomicOdxType) -> AtomicOdxType:
        return physical_value

    def convert_internal_to_physical(self, internal_value: AtomicOdxType) -> AtomicOdxType:
        return internal_value

    def is_valid_physical_value(self, physical_value: AtomicOdxType) -> bool:
        return self.physical_type.isinstance(physical_value)

    def is_valid_internal_value(self, internal_value: AtomicOdxType) -> bool:
        return self.internal_type.isinstance(internal_value)



# SPDX-License-Identifier: MIT
from dataclasses import dataclass, field
from typing import Any
from xml.etree import ElementTree

from typing_extensions import override

from .complexdop import ComplexDop
from .dataobjectproperty import DataObjectProperty
from .decodestate import DecodeState
from .dtcdop import DtcDop
from .encodestate import EncodeState
from .environmentdata import EnvironmentData
from .exceptions import odxraise, odxrequire
from .nameditemlist import NamedItemList
from .odxdoccontext import OdxDocContext
from .odxlink import OdxLinkDatabase, OdxLinkId, OdxLinkRef
from .odxtypes import DataType, ParameterValue, ParameterValueDict
from .parameters.codedconstparameter import CodedConstParameter
from .parameters.parameter import Parameter
from .parameters.parameterwithdop import ParameterWithDOP
from .parameters.physicalconstantparameter import PhysicalConstantParameter
from .parameters.valueparameter import ValueParameter
from .snrefcontext import SnRefContext
from .utils import dataclass_fields_asdict


@dataclass(kw_only=True)
class EnvironmentDataDescription(ComplexDop):
    """This class represents environment data descriptions

    An environment data description provides a list of all environment
    data objects that are potentially applicable to decode a given
    response. (If a given environment data object is applicable
    depends on the value of the DtcDOP that is associated with it.)

    """

    param_snref: str | None = None
    param_snpathref: str | None = None

    # in ODX 2.0.0, ENV-DATAS seems to be a mandatory
    # sub-element of ENV-DATA-DESC, in ODX 2.2 it is not
    # present
    env_datas: NamedItemList[EnvironmentData] = field(default_factory=NamedItemList)
    env_data_refs: list[OdxLinkRef] = field(default_factory=list)

    @staticmethod
    def from_et(et_element: ElementTree.Element,
                context: OdxDocContext) -> "EnvironmentDataDescription":
        """Reads Environment Data Description from Diag Layer."""
        kwargs = dataclass_fields_asdict(ComplexDop.from_et(et_element, context))

        param_snref = None
        if (param_snref_elem := et_element.find("PARAM-SNREF")) is not None:
            param_snref = odxrequire(param_snref_elem.get("SHORT-NAME"))
        param_snpathref = None
        if (param_snpathref_elem := et_element.find("PARAM-SNPATHREF")) is not None:
            param_snpathref = odxrequire(param_snpathref_elem.get("SHORT-NAME-PATH"))

        # ODX 2.0 mandates ENV-DATA-DESC to contain a list of
        # ENV-DATAS and no ENV-DATA-REFS while for ODX 2.2 the
        # situation is reversed. This means that we will create one
        # empty and one non-empty list here. (Which is which depends
        # on the version of the standard used by the file.)
        env_datas = NamedItemList([
            EnvironmentData.from_et(env_data_elem, context)
            for env_data_elem in et_element.iterfind("ENV-DATAS/ENV-DATA")
        ])
        env_data_refs = [
            odxrequire(OdxLinkRef.from_et(env_data_ref, context))
            for env_data_ref in et_element.iterfind("ENV-DATA-REFS/ENV-DATA-REF")
        ]

        return EnvironmentDataDescription(
            param_snref=param_snref,
            param_snpathref=param_snpathref,
            env_datas=env_datas,
            env_data_refs=env_data_refs,
            **kwargs)

    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        odxlinks = {self.odx_id: self}

        if not self.env_data_refs:
            for ed in self.env_datas:
                odxlinks.update(ed._build_odxlinks())

        return odxlinks

    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        # ODX 2.0 specifies environment data objects here, ODX 2.2
        # uses references
        if self.env_data_refs:
            self.env_datas = NamedItemList(
                [odxlinks.resolve(x, EnvironmentData) for x in self.env_data_refs])
        else:
            for ed in self.env_datas:
                ed._resolve_odxlinks(odxlinks)

    def _resolve_snrefs(self, context: SnRefContext) -> None:
        # ODX 2.0 specifies environment data objects here, ODX 2.2
        # uses references
        if self.env_data_refs:
            for ed in self.env_datas:
                ed._resolve_snrefs(context)

    @override
    def encode_into_pdu(self, physical_value: ParameterValue | None,
                        encode_state: EncodeState) -> None:
        """Convert a physical value into bytes and emplace them into a PDU.
        """

        # retrieve the DTC as a numerical value from the referenced
        # parameter (which must be located somewhere before the
        # parameter using the environment data description)
        if self.param_snref is None:
            odxraise("Specifying the DTC parameter for environment data "
                     "descriptions via SNPATHREF is not supported yet")
            return None

        numerical_dtc_value: ParameterValue | None = None
        for prev_param, prev_param_value in reversed(encode_state.journal):
            if prev_param.short_name == self.param_snref:
                numerical_dtc_value = self._get_numerical_dtc_from_parameter(
                    prev_param, prev_param_value)
                break

        if numerical_dtc_value is None:
            odxraise("Environment data description parameters are only allowed following "
                     "the referenced parameter.")
            return

        # deal with the "all value" environment data. This holds
        # parameters that are common to all DTCs. Be aware that the
        # specification mandates that there is at most one such
        # environment data object
        for env_data in self.env_datas:
            if env_data.all_value:
                tmp = encode_state.allow_unknown_parameters
                encode_state.allow_unknown_parameters = True
                env_data.encode_into_pdu(physical_value, encode_state)
                encode_state.allow_unknown_parameters = tmp
                break

        # find the environment data corresponding to the given trouble
        # code
        for env_data in self.env_datas:
            if numerical_dtc_value in env_data.dtc_values:
                tmp = encode_state.allow_unknown_parameters
                encode_state.allow_unknown_parameters = True
                env_data.encode_into_pdu(physical_value, encode_state)
                encode_state.allow_unknown_parameters = tmp
                break

    @override
    def decode_from_pdu(self, decode_state: DecodeState) -> ParameterValue:
        """Extract the bytes from a PDU and convert them to a physical value.
        """

        # retrieve the DTC as a numerical value from the referenced
        # parameter (which must be located somewhere before the
        # parameter using the environment data description)
        if self.param_snref is None:
            odxraise("Specifying the DTC parameter for environment data "
                     "descriptions via SNPATHREF is not supported yet")
            return None

        numerical_dtc_value: ParameterValue | None = None
        for prev_param, prev_param_value in reversed(decode_state.journal):
            if prev_param.short_name == self.param_snref:
                numerical_dtc_value = self._get_numerical_dtc_from_parameter(
                    prev_param, prev_param_value)
                break

        if numerical_dtc_value is None:
            odxraise("Environment data description parameters are only allowed following "
                     "the referenced parameter.")
            return

        result: ParameterValueDict = {}

        # deal with the "all value" environment data. This holds
        # parameters that are common to all DTCs. Be aware that the
        # specification mandates that there is at most one such
        # environment data object
        for env_data in self.env_datas:
            if env_data.all_value:
                tmp = env_data.decode_from_pdu(decode_state)
                if not isinstance(tmp, dict):
                    odxraise()
                result.update(tmp)
                break

        # find the environment data corresponding to the given trouble
        # code
        for env_data in self.env_datas:
            if numerical_dtc_value in env_data.dtc_values:
                tmp = env_data.decode_from_pdu(decode_state)
                if not isinstance(tmp, dict):
                    odxraise()
                result.update(tmp)
                break

        return result

    def _get_numerical_dtc_from_parameter(self, param: Parameter,
                                          param_value: ParameterValue | None) -> int:
        if isinstance(param, ParameterWithDOP):
            dop = param.dop
            if not isinstance(dop, (DataObjectProperty, DtcDop)):
                odxraise(f"The DOP of the parameter referenced by environment data descriptions "
                         f"must use either be DataObjectProperty or a DtcDop (encountered "
                         f"{type(param).__name__} for parameter '{param.short_name}' "
                         f"of ENV-DATA-DESC '{self.short_name}')")
                return 0

            if dop.diag_coded_type.base_data_type != DataType.A_UINT32:
                odxraise(f"The data type used by the DOP of the parameter referenced "
                         f"by environment data descriptions must be A_UINT32 "
                         f"(encountered '{dop.diag_coded_type.base_data_type.value}')")

            if param_value is None:
                if isinstance(param, ValueParameter):
                    param_value = param.physical_default_value
                elif isinstance(param, PhysicalConstantParameter):
                    param_value = param.physical_constant_value
                else:
                    odxraise()  # make mypy happy...
                    return

            if isinstance(dop, DtcDop):
                return dop.convert_to_numerical_trouble_code(odxrequire(param_value))
            elif isinstance(dop, DataObjectProperty):
                return int(dop.compu_method.convert_physical_to_internal(
                    param_value))  # type: ignore[arg-type]

            odxraise()  # not reachable...

        elif isinstance(param, CodedConstParameter):
            if param.diag_coded_type.base_data_type != DataType.A_UINT32:
                odxraise(f"The data type used by the parameter referenced "
                         f"by environment data descriptions must be A_UINT32 "
                         f"(encountered '{param.diag_coded_type.base_data_type.value}')")

                return param.coded_value

            if not isinstance(param.coded_value, int):
                odxraise()

            return param.coded_value

        else:
            odxraise(f"The parameter referenced by environment data descriptions "
                     f"must be a parameter that either specifies a DOP or a constant "
                     f"(encountered {type(param).__name__} for reference '{self.param_snref}' of "
                     f"ENV-DATA-DESC '{self.short_name}')")
            return 0



# SPDX-License-Identifier: MIT
from dataclasses import dataclass
from typing import Any
from xml.etree import ElementTree

from .element import IdentifiableElement
from .odxdoccontext import OdxDocContext
from .odxlink import OdxLinkDatabase, OdxLinkId
from .snrefcontext import SnRefContext
from .utils import dataclass_fields_asdict


@dataclass(kw_only=True)
class PhysicalDimension(IdentifiableElement):
    """A physical dimension is a formal definition of a unit.

    It consists of the exponents for the SI units:

    | Symbol | Name     | Quantity                  | Property                 |
    |  ---   | ---      | ---                       | ---                      |
    | s      | second   | time                      | `time_exp`               |
    | m      | metre    | length                    | `length_exp`             |
    | kg     | kilogram | mass                      | `mass_exp`               |
    | A      | ampere   | electric current          | `current_exp`            |
    | K      | kelvin   | thermodynamic temperature | `temperature_exp`        |
    | mol    | mole     | amount of substance       | `molar_amount_exp`       |
    | cd     | candela  | luminous intensity        | `luminous_intensity_exp` |

    (The first three columns are from https://en.wikipedia.org/wiki/International_System_of_Units.)


    Examples
    --------

    The unit `m/s` (or `m**1 * s**(-1)`) can be represented as
    ```
    PhysicalDimension(
        odx_id="velocity",
        short_name="metre_per_second",
        length_exp=1,
        time_exp=-1
    )
    ```
    """
    length_exp: int | None = None
    mass_exp: int | None = None
    time_exp: int | None = None
    current_exp: int | None = None
    temperature_exp: int | None = None
    molar_amount_exp: int | None = None
    luminous_intensity_exp: int | None = None

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "PhysicalDimension":
        kwargs = dataclass_fields_asdict(IdentifiableElement.from_et(et_element, context))

        def read_optional_int(element: ElementTree.Element, name: str) -> int | None:
            if (val_str := element.findtext(name)) is not None:
                return int(val_str)
            else:
                return None

        length_exp = read_optional_int(et_element, "LENGTH-EXP")
        mass_exp = read_optional_int(et_element, "MASS-EXP")
        time_exp = read_optional_int(et_element, "TIME-EXP")
        current_exp = read_optional_int(et_element, "CURRENT-EXP")
        temperature_exp = read_optional_int(et_element, "TEMPERATURE-EXP")
        molar_amount_exp = read_optional_int(et_element, "MOLAR-AMOUNT-EXP")
        luminous_intensity_exp = read_optional_int(et_element, "LUMINOUS-INTENSITY-EXP")

        return PhysicalDimension(
            length_exp=length_exp,
            mass_exp=mass_exp,
            time_exp=time_exp,
            current_exp=current_exp,
            temperature_exp=temperature_exp,
            molar_amount_exp=molar_amount_exp,
            luminous_intensity_exp=luminous_intensity_exp,
            **kwargs)

    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        return {self.odx_id: self}

    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        pass

    def _resolve_snrefs(self, context: SnRefContext) -> None:
        pass



# SPDX-License-Identifier: MIT
from dataclasses import dataclass
from typing import TYPE_CHECKING
from xml.etree import ElementTree

from .exceptions import odxraise
from .matchingparameter import MatchingParameter
from .odxdoccontext import OdxDocContext

if TYPE_CHECKING:
    from .matchingbasevariantparameter import MatchingBaseVariantParameter
    from .matchingparameter import MatchingParameter


@dataclass(kw_only=True)
class VariantPattern:
    """Variant patterns are used to identify the concrete variant of an ECU

    This is done by observing the responses after sending it some
    requests. There are two kinds of variant patterns:
    `BaseVariantPattern`s which are used to identify the applicable
    base variant and ECU variant patterns which can be used identify
    concrete revisions of the ECU in question. (Both types of pattern
    are optional, i.e., it might not be possible to identify the base-
    or ECU variant present.)
    """

    def get_matching_parameters(
            self) -> list["MatchingParameter"] | list["MatchingBaseVariantParameter"]:
        odxraise(
            f"VariantPattern subclass `{type(self).__name__}` does not "
            f"implement `.get_match_parameters()`", RuntimeError)
        return []

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "VariantPattern":
        return VariantPattern()



# SPDX-License-Identifier: MIT
from dataclasses import dataclass
from typing import Any
from xml.etree import ElementTree

from ..exceptions import odxraise
from ..odxdoccontext import OdxDocContext
from ..odxlink import OdxLinkDatabase, OdxLinkId
from ..odxtypes import AtomicOdxType, DataType
from ..snrefcontext import SnRefContext
from .compucategory import CompuCategory
from .compuinternaltophys import CompuInternalToPhys
from .compuphystointernal import CompuPhysToInternal


@dataclass(kw_only=True)
class CompuMethod:
    """A compu method translates between the internal representation
    of a value and their physical representation.

    There are many compu methods, but all of them are specified using
    the same mechanism: The conversion from internal to physical
    quantities is specified using the COMPU-INTERNAL-TO-PHYS subtag,
    and the inverse is covered by
    COMPU-PHYS-TO-INTERNAL. Alternatively to directly specifying the
    parameters needed for conversion, it is also possible to specify a
    Java program which does the conversion (doing this excludes using
    ODX in non-Java contexts, though).

    For details, refer to ASAM specification MCD-2 D (ODX), section 7.3.6.6.

    """

    category: CompuCategory
    compu_internal_to_phys: CompuInternalToPhys | None = None
    compu_phys_to_internal: CompuPhysToInternal | None = None

    physical_type: DataType
    internal_type: DataType

    @staticmethod
    def compu_method_from_et(et_element: ElementTree.Element, context: OdxDocContext, *,
                             internal_type: DataType, physical_type: DataType) -> "CompuMethod":
        cat_text = et_element.findtext("CATEGORY")
        if cat_text is None:
            odxraise("No category specified for compu method")
            cat_text = "IDENTICAL"

        try:
            category = CompuCategory(cat_text)
        except ValueError:
            odxraise(f"Encountered compu method of unknown category '{cat_text}'")
            category = CompuCategory.IDENTICAL

        compu_internal_to_phys = None
        if (citp_elem := et_element.find("COMPU-INTERNAL-TO-PHYS")) is not None:
            compu_internal_to_phys = CompuInternalToPhys.compu_internal_to_phys_from_et(
                citp_elem, context, internal_type=internal_type, physical_type=physical_type)
        compu_phys_to_internal = None
        if (cpti_elem := et_element.find("COMPU-PHYS-TO-INTERNAL")) is not None:
            compu_phys_to_internal = CompuPhysToInternal.compu_phys_to_internal_from_et(
                cpti_elem, context, internal_type=internal_type, physical_type=physical_type)

        return CompuMethod(
            category=category,
            compu_internal_to_phys=compu_internal_to_phys,
            compu_phys_to_internal=compu_phys_to_internal,
            physical_type=physical_type,
            internal_type=internal_type)

    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        result = {}

        if self.compu_internal_to_phys is not None:
            result.update(self.compu_internal_to_phys._build_odxlinks())

        if self.compu_phys_to_internal is not None:
            result.update(self.compu_phys_to_internal._build_odxlinks())

        return result

    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        if self.compu_internal_to_phys is not None:
            self.compu_internal_to_phys._resolve_odxlinks(odxlinks)

        if self.compu_phys_to_internal is not None:
            self.compu_phys_to_internal._resolve_odxlinks(odxlinks)

    def _resolve_snrefs(self, context: SnRefContext) -> None:
        if self.compu_internal_to_phys is not None:
            self.compu_internal_to_phys._resolve_snrefs(context)

        if self.compu_phys_to_internal is not None:
            self.compu_phys_to_internal._resolve_snrefs(context)

    def convert_physical_to_internal(self, physical_value: AtomicOdxType) -> AtomicOdxType:
        raise NotImplementedError()

    def convert_internal_to_physical(self, internal_value: AtomicOdxType) -> AtomicOdxType:
        raise NotImplementedError()

    def is_valid_physical_value(self, physical_value: AtomicOdxType) -> bool:
        raise NotImplementedError()

    def is_valid_internal_value(self, internal_value: AtomicOdxType) -> bool:
        raise NotImplementedError()



# SPDX-License-Identifier: MIT
from dataclasses import dataclass, field
from typing import Any, Literal
from xml.etree import ElementTree

from typing_extensions import final, override

from ..decodestate import DecodeState
from ..element import NamedElement
from ..encodestate import EncodeState
from ..odxdoccontext import OdxDocContext
from ..odxlink import OdxLinkDatabase, OdxLinkId
from ..odxtypes import ParameterValue
from ..snrefcontext import SnRefContext
from ..specialdatagroup import SpecialDataGroup
from ..utils import dataclass_fields_asdict

ParameterType = Literal[
    "CODED-CONST",
    "DYNAMIC",
    "LENGTH-KEY",
    "MATCHING-REQUEST-PARAM",
    "NRC-CONST",
    "PHYS-CONST",
    "RESERVED",
    "SYSTEM",
    "TABLE-ENTRY",
    "TABLE-KEY",
    "TABLE-STRUCT",
    "VALUE",
]


@dataclass(kw_only=True)
class Parameter(NamedElement):
    """This class corresponds to POSITIONABLE-PARAM in the ODX
    specification

    All parameter classes must adhere to the `Codec` type protocol, so
    `isinstance(param, Codec)` ought to be true.  Be aware that, even
    though the ODX specification seems to make the distinction of
    "positionable" and "normal" parameters, it does not define any
    non-positionable parameter types.

    """
    sdgs: list[SpecialDataGroup] = field(default_factory=list)
    semantic: str | None = None
    oid: str | None = None
    byte_position: int | None = None
    bit_position: int | None = None

    @property
    def parameter_type(self) -> ParameterType:
        raise NotImplementedError(
            ".parameter_type is not implemented by the concrete parameter class")

    @property
    def is_required(self) -> bool:
        """True if the parameter must be explicitly specified when
        encoding a message.

        Required parameters are always settable, and parameters which
        have a default value are settable but not required to be
        specified.

        """
        raise NotImplementedError(".is_required is not implemented by the concrete parameter class")

    @property
    def is_settable(self) -> bool:
        """True if the parameter can be specified when encoding a
        message.

        Required parameters are always settable, and parameters which
        have a default value are settable but not required to be
        specified.
        """
        raise NotImplementedError(".is_settable is not implemented by the concrete parameter class")

    @staticmethod
    @override
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "Parameter":

        kwargs = dataclass_fields_asdict(NamedElement.from_et(et_element, context))

        sdgs = [SpecialDataGroup.from_et(sdge, context) for sdge in et_element.iterfind("SDGS/SDG")]
        semantic = et_element.attrib.get("SEMANTIC")
        oid = et_element.attrib.get("OID")
        byte_position_str = et_element.findtext("BYTE-POSITION")
        byte_position = int(byte_position_str) if byte_position_str is not None else None
        bit_position_str = et_element.findtext("BIT-POSITION")
        bit_position = int(bit_position_str) if bit_position_str is not None else None

        return Parameter(
            sdgs=sdgs,
            semantic=semantic,
            oid=oid,
            byte_position=byte_position,
            bit_position=bit_position,
            **kwargs)

    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        result = {}

        for sdg in self.sdgs:
            result.update(sdg._build_odxlinks())

        return result

    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        for sdg in self.sdgs:
            sdg._resolve_odxlinks(odxlinks)

    def _resolve_snrefs(self, context: SnRefContext) -> None:
        for sdg in self.sdgs:
            sdg._resolve_snrefs(context)

    def get_static_bit_length(self) -> int | None:
        return None

    @final
    def encode_into_pdu(self, physical_value: ParameterValue | None,
                        encode_state: EncodeState) -> None:
        """Convert a physical value into its encoded form and place it
        into the PDU

        Also, adapt the `encode_state` so that it points to where the
        next parameter is located (if the next parameter does not
        explicitly specify a position)

        """

        if self.byte_position is not None:
            encode_state.cursor_byte_position = encode_state.origin_byte_position + self.byte_position

        encode_state.cursor_bit_position = self.bit_position or 0

        self._encode_positioned_into_pdu(physical_value, encode_state)

        encode_state.cursor_bit_position = 0

    def _encode_positioned_into_pdu(self, physical_value: ParameterValue | None,
                                    encode_state: EncodeState) -> None:
        """Method which actually encodes the parameter

        Its location is managed by `Parameter`."""
        raise NotImplementedError(
            f"Required method '_encode_positioned_into_pdu()' not implemented by "
            f"child class {type(self).__name__}")

    @final
    def decode_from_pdu(self, decode_state: DecodeState) -> ParameterValue:
        """Retrieve the raw data for the parameter from the PDU and
        convert it to its physical interpretation

        Also, adapt the `encode_state` so that it points to where the
        next parameter is located (if the next parameter does not
        explicitly specify a position)

        """

        if self.byte_position is not None:
            decode_state.cursor_byte_position = decode_state.origin_byte_position + self.byte_position

        decode_state.cursor_bit_position = self.bit_position or 0

        result = self._decode_positioned_from_pdu(decode_state)

        decode_state.cursor_bit_position = 0

        return result

    def _decode_positioned_from_pdu(self, decode_state: DecodeState) -> ParameterValue:
        """Method which actually decodes the parameter

        Its location is managed by `Parameter`."""
        raise NotImplementedError(
            f"Required method '_decode_positioned_from_pdu()' not implemented by "
            f"child class {type(self).__name__}")



# SPDX-License-Identifier: MIT
from dataclasses import dataclass
from xml.etree import ElementTree

from .exceptions import odxrequire
from .odxdoccontext import OdxDocContext
from .utils import read_hex_binary


# note that the spec has a typo here: it calls the corresponding
# XML tag POS-RESPONSE-SUPPRESSABLE...
@dataclass(kw_only=True)
class PosResponseSuppressible:
    bit_mask: int

    coded_const_snref: str | None = None
    coded_const_snpathref: str | None = None

    value_snref: str | None = None
    value_snpathref: str | None = None

    phys_const_snref: str | None = None
    phys_const_snpathref: str | None = None

    table_key_snref: str | None = None
    table_key_snpathref: str | None = None

    @staticmethod
    def from_et(et_element: ElementTree.Element,
                context: OdxDocContext) -> "PosResponseSuppressible":

        bit_mask = odxrequire(read_hex_binary(et_element.find("BIT-MASK")))

        coded_const_snref = None
        if (cc_snref_elem := et_element.find("CODED-CONST-SNREF")) is not None:
            coded_const_snref = cc_snref_elem.attrib["SHORT-NAME"]
        coded_const_snpathref = None
        if (cc_snpathref_elem := et_element.find("CODED-CONST-SNPATHREF")) is not None:
            coded_const_snpathref = cc_snpathref_elem.attrib["SHORT-NAME-PATH"]

        value_snref = None
        if (cc_snref_elem := et_element.find("VALUE-SNREF")) is not None:
            value_snref = cc_snref_elem.attrib["SHORT-NAME"]
        value_snpathref = None
        if (cc_snpathref_elem := et_element.find("VALUE-SNPATHREF")) is not None:
            value_snpathref = cc_snpathref_elem.attrib["SHORT-NAME-PATH"]

        phys_const_snref = None
        if (cc_snref_elem := et_element.find("PHYS-CONST-SNREF")) is not None:
            phys_const_snref = cc_snref_elem.attrib["SHORT-NAME"]
        phys_const_snpathref = None
        if (cc_snpathref_elem := et_element.find("PHYS-CONST-SNPATHREF")) is not None:
            phys_const_snpathref = cc_snpathref_elem.attrib["SHORT-NAME-PATH"]

        table_key_snref = None
        if (cc_snref_elem := et_element.find("TABLE-KEY-SNREF")) is not None:
            table_key_snref = cc_snref_elem.attrib["SHORT-NAME"]
        table_key_snpathref = None
        if (cc_snpathref_elem := et_element.find("TABLE-KEY-SNPATHREF")) is not None:
            table_key_snpathref = cc_snpathref_elem.attrib["SHORT-NAME-PATH"]

        return PosResponseSuppressible(
            bit_mask=bit_mask,
            coded_const_snref=coded_const_snref,
            coded_const_snpathref=coded_const_snpathref,
            value_snref=value_snref,
            value_snpathref=value_snpathref,
            phys_const_snref=phys_const_snref,
            phys_const_snpathref=phys_const_snpathref,
            table_key_snref=table_key_snref,
            table_key_snpathref=table_key_snpathref,
        )



# SPDX-License-Identifier: MIT
from typing import TYPE_CHECKING, NoReturn, TypeVar

from .globals import logger


class OdxError(Exception):
    """Any error that happens during interacting with diagnostic objects."""


class EncodeError(Warning, OdxError):
    """Encoding of a message to raw data failed"""


class DecodeError(Warning, OdxError):
    """Decoding raw data failed."""


class DecodeMismatch(DecodeError):
    """Decoding failed because some parameters exhibit an incorrect value

    This is can happen if NRC-CONST or environment data descriptions
    are present.

    """


class OdxWarning(Warning):
    """Any warning that happens during interacting with diagnostic objects."""


#: Specify whether violations of the ODX specification or failed
#: assumptions in the ODX library code should cause the routine in
#: question to abort or whether it should continue anyway. Be aware that
#: in non-strict mode, the behavior of odxtools is undefined and it
#: might start eating children...
strict_mode = True


def odxraise(message: str | None = None, error_type: type[Exception] = OdxError) -> NoReturn:
    """
    Raise an exception but only if in strict mode.

    Also, convince type checkers that the exception is always raised.
    """
    if TYPE_CHECKING or strict_mode:
        if message is None:
            raise error_type()
        else:
            raise error_type(message)
    elif message is not None:
        logger.warning(message)


def odxassert(condition: bool,
              message: str | None = None,
              error_type: type[Exception] = OdxError) -> None:
    """
    This method works similar as the build-in `assert` statement

    The differences are that instead of raising an `AssertationError`,
    an `OdxError` is raised by default and that is possible to not
    raise any exception by setting `odxtools.exceptions.strict_mode` to
    `False`. The latter is convenient when having to deal with files
    that do not comply to the ODX specification, but it may lead to
    undefined behavior. (Use the non-strict mode with great care.)
    """
    if not condition:
        odxraise(message, error_type)


T = TypeVar("T")


def odxrequire(obj: T | None, message: str | None = None) -> T:
    """This function ensures that an object required by the ODX
    specification is actually present.

    In other words, that the object in question is not `None`. If such
    a mandatory object is not present, an `OdxError` is raised if the
    `odxtools.exceptions.strict_mode` flag is `True`. (Note that this
    function is pretty useful for type checking.)
    """
    if obj is None:
        odxraise(message)

    return obj



# SPDX-License-Identifier: MIT
from dataclasses import dataclass, field
from typing import Any, Optional
from xml.etree import ElementTree

from .companydocinfo import CompanyDocInfo
from .docrevision import DocRevision
from .odxdoccontext import OdxDocContext
from .odxlink import OdxLinkDatabase, OdxLinkId
from .snrefcontext import SnRefContext


@dataclass(kw_only=True)
class AdminData:
    language: str | None = None
    company_doc_infos: list[CompanyDocInfo] = field(default_factory=list)
    doc_revisions: list[DocRevision] = field(default_factory=list)

    @staticmethod
    def from_et(et_element: ElementTree.Element | None,
                context: OdxDocContext) -> Optional["AdminData"]:

        if et_element is None:
            return None

        language = et_element.findtext("LANGUAGE")

        company_doc_infos = [
            CompanyDocInfo.from_et(cdi_elem, context)
            for cdi_elem in et_element.iterfind("COMPANY-DOC-INFOS/COMPANY-DOC-INFO")
        ]

        doc_revisions = [
            DocRevision.from_et(dr_elem, context)
            for dr_elem in et_element.iterfind("DOC-REVISIONS/DOC-REVISION")
        ]

        return AdminData(
            language=language, company_doc_infos=company_doc_infos, doc_revisions=doc_revisions)

    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        result: dict[OdxLinkId, Any] = {}

        for cdi in self.company_doc_infos:
            result.update(cdi._build_odxlinks())

        for dr in self.doc_revisions:

            result.update(dr._build_odxlinks())

        return result

    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        for cdi in self.company_doc_infos:
            cdi._resolve_odxlinks(odxlinks)

        skipped_ref_ids : set[str] = set()
        for dr in self.doc_revisions:
            if dr.team_member_ref:
                if "_Porsche_AG" not in dr.team_member_ref.ref_id:
                    dr._resolve_odxlinks(odxlinks)
                elif  dr.team_member_ref.ref_id not in skipped_ref_ids:
                    print("[Warning] odxtools_porsche parser skipped team_ref_resolution: ", dr.team_member_ref.ref_id)
                    skipped_ref_ids.add(dr.team_member_ref.ref_id)

    def _resolve_snrefs(self, context: SnRefContext) -> None:
        for cdi in self.company_doc_infos:
            cdi._resolve_snrefs(context)

        for dr in self.doc_revisions:
            dr._resolve_snrefs(context)



# SPDX-License-Identifier: MIT
from enum import Enum


class DiagLayerType(Enum):
    PROTOCOL = "PROTOCOL"
    FUNCTIONAL_GROUP = "FUNCTIONAL-GROUP"
    BASE_VARIANT = "BASE-VARIANT"
    ECU_VARIANT = "ECU-VARIANT"
    ECU_SHARED_DATA = "ECU-SHARED-DATA"

    @property
    def inheritance_priority(self) -> int:
        """Return the inheritance priority of diag layers of the given type

        ODX mandates that diagnostic layers can only inherit from
        layers of lower priority...

        """

        PRIORITY_OF_DIAG_LAYER_TYPE: dict[DiagLayerType, int] = {
            DiagLayerType.PROTOCOL:
                1,
            DiagLayerType.FUNCTIONAL_GROUP:
                2,
            DiagLayerType.BASE_VARIANT:
                3,
            DiagLayerType.ECU_VARIANT:
                4,

            # ECU shared data layers are a bit weird (see section
            # 7.3.2.4.4 of the ASAM specification): they can be
            # inherited from by any other layer but they will
            # override any objects which are also provided by any of
            # the other parent layers. tl;dr: When it comes to
            # inheritance, they have the highest priority.
            DiagLayerType.ECU_SHARED_DATA:
                100,
        }

        return PRIORITY_OF_DIAG_LAYER_TYPE[self]



#! /usr/bin/python3
#
# SPDX-License-Identifier: MIT
import argparse
from typing import Any, Protocol, runtime_checkable

from ..database import Database
from ..loadfile import load_file as _load_file


@runtime_checkable
class SubparsersList(Protocol):
    """This protocol reproduces the parts of `argparse._SubparsersAction` which are needed by odxtools

    Unfortunately this is necessary because
    `argparse._SubparsersAction` is a private class of the `argparse`
    module that is used by some of its public API and thus cannot be
    used outside of the `argparse` module.
    """

    def add_parser(self, name: str, **kwargs: Any) -> "argparse.ArgumentParser":
        ...


def add_pdx_argument(parser: argparse.ArgumentParser, is_optional: bool = False) -> None:
    parser.add_argument(
        "pdx_file",
        help="Location of the .pdx file",
        nargs="?" if is_optional else 1,
        metavar="PDX_FILE")


def load_file(args: argparse.Namespace) -> Database:
    pdx_file_name = args.pdx_file if isinstance(args.pdx_file, str) else args.pdx_file[0]
    return _load_file(pdx_file_name)
